import sys
import os
import json
from typing import Any, Dict, List, Optional
from pydantic import BaseModel

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from core.problem_spec import ProblemSpec
from pipeline import UniversalSolver
from interface.nl_parser import parse_description, extract_data_from_description

_FASTAPI_AVAILABLE = False
try:
    from fastapi import FastAPI, HTTPException
    import uvicorn
    _FASTAPI_AVAILABLE = True
except ImportError:
    FastAPI = None
    HTTPException = None
    uvicorn = None


class SolveRequest(BaseModel):
    name: str = ""
    problem_type: str = "transformation"
    inputs: Dict = {"data": {"type": "list"}}
    outputs: Dict = {"result": {"type": "list"}}
    constraints: List[str] = []
    objectives: List[Dict] = []
    data: Any = None
    use_llm: bool = False


class SolveNLRequest(BaseModel):
    description: str
    data: Any = None
    use_llm: bool = False


class ExplainRequest(BaseModel):
    algorithm_name: str
    detail: str = "short"


def create_app(solver: Optional[UniversalSolver] = None):
    if not _FASTAPI_AVAILABLE:
        return None

    if solver is None:
        solver = UniversalSolver()

    app = FastAPI(title="AAlgoI API", version="1.0.0",
                  description="Universal Problem-Solving System API")

    @app.post("/solve")
    async def solve(request: SolveRequest):
        try:
            spec = ProblemSpec(
                name=request.name or "api_problem",
                inputs=request.inputs,
                outputs=request.outputs,
                constraints=request.constraints,
                objectives=request.objectives
            )
            data = request.data if request.data is not None else [3, 1, 2]
            result = solver.solve(spec, data, use_llm=request.use_llm)

            return {
                "success": result["success"],
                "result": _serialize_result(result["result"]),
                "time_ms": result["time_ms"],
                "pipeline": result["pipeline"],
                "strategy": result.get("selection", {}).get("synthesis_strategy", ""),
                "confidence": result.get("selection", {}).get("confidence", 0),
                "validation": result["validation"],
                "explanation": [
                    {"algorithm": e.algorithm_name, "summary": e.summary[:200]}
                    for e in result.get("explanation", [])
                ]
            }
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/solve/nl")
    async def solve_natural_language(request: SolveNLRequest):
        try:
            spec = parse_description(request.description)
            data = request.data if request.data is not None else extract_data_from_description(request.description)
            if data is None:
                data = [3, 1, 4, 1, 5]

            result = solver.solve(spec, data, use_llm=request.use_llm)

            return {
                "success": result["success"],
                "result": _serialize_result(result["result"]),
                "time_ms": result["time_ms"],
                "pipeline": result["pipeline"],
                "strategy": result.get("selection", {}).get("synthesis_strategy", ""),
                "confidence": result.get("selection", {}).get("confidence", 0),
                "parsed_spec": {
                    "name": spec.name,
                    "problem_type": spec.problem_type.value,
                    "inputs": spec.inputs,
                    "outputs": spec.outputs
                },
                "validation": result["validation"]
            }
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/stats")
    async def get_stats():
        try:
            stats = solver.get_stats()
            return stats
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/explain")
    async def explain_algorithm(request: ExplainRequest):
        from core.explainer import Explainer
        explainer = Explainer()
        exp = explainer.explain(request.algorithm_name, detail=request.detail)
        return {
            "algorithm": exp.algorithm_name,
            "summary": exp.summary,
            "complexity": exp.complexity,
            "steps": exp.steps,
            "best_for": exp.best_for
        }

    @app.get("/health")
    async def health():
        return {"status": "ok", "service": "aalgoi-api"}

    return app


def _serialize_result(result: Any) -> Any:
    if isinstance(result, (int, float, str, bool)):
        return result
    if isinstance(result, (list, tuple)):
        return [_serialize_result(r) for r in result]
    if isinstance(result, dict):
        return {str(k): _serialize_result(v) for k, v in result.items()}
    return str(result)


def run(host="0.0.0.0", port=8000, solver=None):
    if not _FASTAPI_AVAILABLE:
        print("FastAPI not available. Install with: pip install fastapi uvicorn")
        return
    app = create_app(solver)
    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    run()




#!/usr/bin/env python3
"""
AAlgoI - Artificial Algorithm Intelligence
Main entry point for the adaptive algorithm system.

Usage:
    python main.py --mode demo
    python main.py --mode benchmark --size 10000
    python main.py --mode interactive
    python main.py --mode learning --llm-enabled --llm-model phi3:mini
"""

import argparse
import random
import time
import json
from typing import List, Dict, Any

from pipeline import AAlgoI

def generate_test_data(size: int, pattern: str = "random") -> List[int]:
    if pattern == "random":
        return [random.randint(0, size * 10) for _ in range(size)]
    elif pattern == "sorted":
        return list(range(size))
    elif pattern == "reverse":
        return list(range(size, 0, -1))
    elif pattern == "nearly_sorted":
        data = list(range(size))
        swaps = size // 20
        for _ in range(swaps):
            i, j = random.randint(0, size-1), random.randint(0, size-1)
            data[i], data[j] = data[j], data[i]
        return data
    elif pattern == "few_unique":
        return [random.randint(0, 10) for _ in range(size)]
    else:
        return [random.randint(0, size * 10) for _ in range(size)]

def demo_mode(config: Dict = None):
    print("=" * 70)
    print("  AAlgoI - Artificial Algorithm Intelligence v2.0")
    print("  Adaptive Algorithm Selection & Optimization Demo")
    print("=" * 70)
    print()

    system = AAlgoI(config={
        "strategy": "hybrid",
        "time_budget_ms": 500,
        "priority": "balanced",
        **(config or {})
    })

    test_cases = [
        ("tiny_random", 15, "random"),
        ("small_sorted", 100, "sorted"),
        ("medium_random", 5000, "random"),
        ("large_random", 50000, "random"),
        ("nearly_sorted", 10000, "nearly_sorted"),
        ("reverse_sorted", 5000, "reverse"),
    ]

    print("Running adaptive algorithm selection on various datasets...")
    print("-" * 70)

    for name, size, pattern in test_cases:
        data = generate_test_data(size, pattern)
        expected = sorted(data)

        print(f"\n  Test: {name} (size={size}, pattern={pattern})")

        result = system.run(data, task_type="sorting", expected_result=expected)

        stats = system.get_stats()
        pipeline = stats["active_pipeline"]

        if system.tracker.measurements:
            last_metric = system.tracker.measurements[-1]["metrics"]
            print(f"   Selected: {' -> '.join(pipeline)}")
            print(f"   Time: {last_metric['wall_time_ms']:.2f}ms | "
                  f"Quality: {last_metric['quality_score']:.2f} | "
                  f"Success: {last_metric['success']}")

        explanation = system.explain_decision()
        for reason in explanation.get("reasoning", [])[:3]:
            print(f"   {reason}")

    print()
    print("-" * 70)
    print("\n  System Statistics:")
    stats = system.get_stats()
    print(f"   Total executions: {stats['executions']}")
    print(f"   Average time: {stats['avg_time_ms']:.2f}ms")
    print(f"   Knowledge base records: {stats['knowledge_base']['total_records']}")
    print(f"   Meta-controller strategy: {stats['meta_controller']['strategy']}")
    print(f"   History size: {stats['meta_controller']['history_size']}")
    print(f"   Hot cache entries: {stats['hot_cache_size']}")
    print(f"   Drift events: {stats['drift']['drift_count']}")
    print(f"   Validation failures: {stats['validator']['total_failures']}")
    if stats['meta_controller']['llm']['enabled']:
        print(f"   LLM: {stats['meta_controller']['llm']['model']} ({'available' if stats['meta_controller']['llm']['available'] else 'unavailable'})")

    print()
    print("\n  Algorithm Performance Summary:")
    for algo_name, summary in stats["performance_tracker"]["summaries"].items():
        if summary:
            print(f"   {algo_name:15s}: avg={summary.get('avg_time_ms', 0):8.2f}ms, "
                  f"quality={summary.get('avg_quality', 0):.2f}, "
                  f"success_rate={summary.get('success_rate', 0):.1%}")

def benchmark_mode(size: int = 10000, config: Dict = None):
    print("=" * 70)
    print(f"  AAlgoI Benchmark - Size: {size}")
    print("=" * 70)
    print()

    system = AAlgoI(config=config or {})

    patterns = ["random", "sorted", "nearly_sorted", "reverse", "few_unique"]

    for pattern in patterns:
        print(f"\n  Pattern: {pattern}")
        data = generate_test_data(size, pattern)

        results = system.benchmark(data)

        sorted_results = sorted(
            results.items(),
            key=lambda x: x[1]["time_ms"] if x[1]["success"] else float('inf')
        )

        for algo_name, result in sorted_results:
            status = "+" if result["success"] and result["correct"] else "-"
            print(f"   {status} {algo_name:15s}: {result['time_ms']:8.2f}ms")

        result = system.run(data, task_type="sorting")
        chosen = system.get_stats()["active_pipeline"]
        print(f"   * AAlgoI chose: {' -> '.join(chosen)}")

def learning_demo(config: Dict = None):
    print("=" * 70)
    print("  AAlgoI Learning Demo v2.0")
    print("  Showing how the system improves over time with bandit + vector KB")
    print("=" * 70)
    print()

    system = AAlgoI(config={
        "strategy": "hybrid",
        "retrain_interval": 50,
        **(config or {})
    })

    print("Training phase: Running 200 executions with varying data...")

    for i in range(200):
        size = random.randint(100, 10000)
        pattern = random.choice(["random", "sorted", "nearly_sorted", "reverse"])
        data = generate_test_data(size, pattern)
        expected = sorted(data)

        system.run(data, task_type="sorting", expected_result=expected)

        if (i + 1) % 50 == 0:
            stats = system.get_stats()
            print(f"  Completed {i+1} executions... (confidence: {stats['meta_controller']['last_confidence']:.2f})")

    print()
    print("Testing phase: Evaluating on new data...")

    test_cases = [
        ("small_random", 50, "random"),
        ("medium_sorted", 5000, "sorted"),
        ("large_nearly_sorted", 20000, "nearly_sorted"),
    ]

    for name, size, pattern in test_cases:
        data = generate_test_data(size, pattern)
        expected = sorted(data)

        result = system.run(data, task_type="sorting", expected_result=expected)

        stats = system.get_stats()
        pipeline = stats["active_pipeline"]

        if system.tracker.measurements:
            last_metric = system.tracker.measurements[-1]["metrics"]
            print(f"\n{name}: {' -> '.join(pipeline)}")
            print(f"  Time: {last_metric['wall_time_ms']:.2f}ms, "
                  f"Quality: {last_metric['quality_score']:.2f}, "
                  f"Confidence: {stats['meta_controller']['last_confidence']:.0%}")

    print()
    print("\n  Final Knowledge Base Stats:")
    kb_stats = system.knowledge.get_all_stats()
    for algo, stats in kb_stats.items():
        print(f"  {algo}: {stats['total_executions']} runs, "
              f"avg_time={stats['avg_time_ms']:.2f}ms, "
              f"success={stats['success_rate']:.1%}, "
              f"score={stats.get('avg_score', 0):.2f}")

    print()
    print("  Bandit Stats:")
    bandit_stats = system.meta_controller.bandit.get_stats()
    print(f"  Epsilon: {bandit_stats['epsilon']:.3f}")
    print(f"  Total trials: {bandit_stats['total_trials']}")
    for name, bstats in bandit_stats['algorithm_stats'].items():
        if bstats['count'] > 0:
            print(f"  {name:15s}: n={bstats['count']:4d}, avg_reward={bstats['avg_reward']:.3f}")

def main():
    parser = argparse.ArgumentParser(description="AAlgoI - Artificial Algorithm Intelligence v2.0")
    parser.add_argument("--mode", choices=["demo", "benchmark", "interactive", "learning"],
                       default="demo", help="Execution mode")
    parser.add_argument("--size", type=int, default=10000, help="Data size for benchmark")
    parser.add_argument("--strategy", choices=["rule-based", "ml-based", "genetic", "hybrid"],
                       default="hybrid", help="Meta-controller strategy")
    parser.add_argument("--llm-enabled", action="store_true", help="Enable LLM advisor")
    parser.add_argument("--llm-backend", choices=["ollama", "llamacpp"], default="ollama",
                       help="LLM backend")
    parser.add_argument("--llm-model", default="phi3:mini", help="LLM model name")
    parser.add_argument("--priority", choices=["speed", "accuracy", "balanced"],
                       default="balanced", help="Execution priority")
    parser.add_argument("--time-budget", type=int, default=500, help="Time budget in ms")
    parser.add_argument("--no-dag", action="store_true", help="Disable DAG pipeline")

    args = parser.parse_args()

    config = {}
    config["strategy"] = args.strategy
    config["priority"] = args.priority
    config["time_budget_ms"] = args.time_budget
    config["enable_dag"] = not args.no_dag
    config["llm"] = {
        "enabled": args.llm_enabled,
        "backend": args.llm_backend,
        "model": args.llm_model
    }

    if args.mode == "demo":
        demo_mode(config)
    elif args.mode == "benchmark":
        benchmark_mode(args.size, config)
    elif args.mode == "interactive":
        print("Interactive mode not available in this environment")
    elif args.mode == "learning":
        learning_demo(config)

if __name__ == "__main__":
    main()



import sys, os, json, time, random
sys.path.insert(0, ".")
import numpy as np
import torch

from core.rl import PPOAgent
from core.context_engine import ContextEngine
from core.problem_spec import ProblemSpec, ProblemType
from pipeline import UniversalSolver

solver = UniversalSolver()
context_engine = ContextEngine()
registry = solver._get_global_registry()
algo_names = list(registry.keys())
algo_to_idx = {n: i for i, n in enumerate(algo_names)}
idx_to_algo = {i: n for n, i in algo_to_idx.items()}

def build_state(data, problem_type):
    context = context_engine.analyze(data, problem_type.value)
    vec = np.zeros(200, dtype=np.float32)
    dp = context.get("data_profile", {})
    patterns = dp.get("patterns", {})
    vec[0] = np.log10(dp.get("size", 1) + 1) / 5.0
    vec[1] = 1.0 if patterns.get("is_sorted") else 0.0
    vec[2] = 1.0 if patterns.get("is_nearly_sorted") else 0.0
    env_info = context.get("environment", {})
    cpu = env_info.get("cpu", {})
    mem = env_info.get("memory", {})
    vec[10] = (100.0 - cpu.get("percent_used", 50)) / 100.0 if cpu else 0.5
    vec[11] = mem.get("available_gb", 1) / max(mem.get("total_gb", 1), 1)
    all_types = list(ProblemType)
    ptype = problem_type
    if ptype in all_types:
        idx = all_types.index(ptype)
        if 20 + idx < 200:
            vec[20 + idx] = 1.0
    return vec

def eval_algo(algo_name, data):
    algo = registry[algo_name]
    try:
        start = time.time()
        result = algo.process(data)
        elapsed = time.time() - start
        valid = algo.validate_output(data, result)
        return valid, elapsed * 1000, result
    except Exception as e:
        return False, 0, str(e)

def run_domain_test(domain, test_cases):
    print(f"\n  {'='*60}")
    print(f"  Domain: {domain.upper()}")
    print(f"  {'='*60}")
    for spec, data, desc in test_cases:
        state = build_state(data, spec.problem_type)
        action, log_prob, value = agent.select_action(state, deterministic=True)
        algo_name = idx_to_algo[action]
        valid, ms, _ = eval_algo(algo_name, data)
        status = "PASS" if valid else "FAIL"
        print(f"    [{status}] {desc:40s} -> {algo_name:20s} ({ms:6.1f}ms)")

def generate_graph(num_nodes, density=0.3):
    nodes = [chr(ord('A') + i) if i < 26 else f"N{i}" for i in range(num_nodes)]
    graph = {n: {} for n in nodes}
    for i in range(num_nodes):
        for j in range(num_nodes):
            if i != j and random.random() < density:
                weight = random.randint(1, 20)
                graph[nodes[i]][nodes[j]] = weight
    return {"graph": graph, "start": nodes[0], "goal": nodes[-1]}

def generate_knapsack(num_items):
    items = [(random.randint(1, 50), random.randint(1, 100)) for _ in range(num_items)]
    capacity = sum(w for w, _ in items) // 2
    return {"items": items, "capacity": capacity}

for label, ckpt_path in [
    ("Final Model", "checkpoints/pretrain_v2/pretrained_final.pt"),
    ("Best Ep1500", "checkpoints/pretrain_v2/best_model_ep1500.pt"),
    ("Best Ep1000", "checkpoints/pretrain_v2/best_model_ep1000.pt"),
]:
    agent = PPOAgent(state_dim=200, num_actions=len(algo_names), config={})
    agent.load(ckpt_path)
    agent.network.eval()
    
    print(f"\n{'='*70}")
    print(f"  {label}")
    print(f"  {'='*70}")
    
    run_domain_test("sorting", [
        (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
         [random.randint(0, 1000) for _ in range(100)], "100 random ints"),
        (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
         list(range(1000)), "1K sorted"),
        (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
         list(range(10000, 0, -1)), "10K reverse"),
        (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
         [random.randint(0, 10) for _ in range(50)], "few unique (50)"),
        (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
         [random.randint(0, 100000) for _ in range(5000)], "5K random"),
    ])
    
    run_domain_test("pathfinding", [
        (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
         generate_graph(10, 0.4), "10-node sparse graph"),
        (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
         generate_graph(26, 0.3), "26-node graph"),
        (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
         generate_graph(50, 0.15), "50-node sparse graph"),
        (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
         generate_graph(20, 0.6), "20-node dense graph"),
    ])
    
    run_domain_test("optimization", [
        (ProblemSpec(name="o", problem_type=ProblemType.OPTIMIZATION),
         generate_knapsack(10), "10-item knapsack"),
        (ProblemSpec(name="o", problem_type=ProblemType.OPTIMIZATION),
         generate_knapsack(20), "20-item knapsack"),
        (ProblemSpec(name="o", problem_type=ProblemType.OPTIMIZATION),
         generate_knapsack(50), "50-item knapsack"),
    ])

print(f"\n{'='*70}")
print("  All tests complete!")
print(f"{'='*70}")




#!/usr/bin/env python3
"""
AAlgoI - Example Usage Script v2.0

Demonstrates all new features: vector KB, validation, bandit, drift detection,
DAG pipelines, LLM advisor, decision audit log, and genetic evolution.
"""

from pipeline import AAlgoI
import random
import time
import numpy as np


def example_1_basic_sorting():
    """Basic sorting with automatic algorithm selection"""
    print("=" * 70)
    print("Example 1: Basic Sorting")
    print("=" * 70)

    system = AAlgoI(config={
        "strategy": "hybrid",
        "time_budget_ms": 500
    })

    datasets = [
        ("Tiny random", [random.randint(0, 100) for _ in range(10)]),
        ("Small sorted", list(range(100))),
        ("Medium random", [random.randint(0, 10000) for _ in range(5000)]),
        ("Nearly sorted", _make_nearly_sorted(10000)),
        ("Large random", [random.randint(0, 100000) for _ in range(50000)]),
    ]

    for name, data in datasets:
        expected = sorted(data)
        result = system.run(data, task_type="sorting", expected_result=expected)

        pipeline = system.get_stats()["active_pipeline"]
        last_metric = system.tracker.measurements[-1]["metrics"]

        print(f"\n{name}:")
        print(f"  Selected: {' -> '.join(pipeline)}")
        print(f"  Time: {last_metric['wall_time_ms']:.2f}ms")
        print(f"  Quality: {last_metric['quality_score']:.2f}")
        print(f"  Correct: {result == expected}")


def example_2_strategy_comparison():
    """Compare different selection strategies"""
    print("\n" + "=" * 70)
    print("Example 2: Strategy Comparison (with bandit + confidence)")
    print("=" * 70)

    strategies = ["rule-based", "hybrid"]
    data = [random.randint(0, 10000) for _ in range(10000)]
    expected = sorted(data)

    for strategy in strategies:
        system = AAlgoI(config={"strategy": strategy})

        start = time.perf_counter()
        result = system.run(data, task_type="sorting", expected_result=expected)
        elapsed = (time.perf_counter() - start) * 1000

        stats = system.get_stats()
        mc_stats = stats["meta_controller"]
        pipeline = stats["active_pipeline"]
        print(f"\n{strategy:15s}: {' -> '.join(pipeline):20s} | "
              f"{elapsed:8.2f}ms | confidence={mc_stats['last_confidence']:.0%}")


def example_3_learning_over_time():
    """Show how system learns and improves with bandit exploration"""
    print("\n" + "=" * 70)
    print("Example 3: Learning Over Time (with bandit evolution)")
    print("=" * 70)

    system = AAlgoI(config={
        "strategy": "hybrid",
        "retrain_interval": 50
    })

    print("\nTraining phase (100 executions)...")
    for i in range(100):
        size = random.randint(100, 5000)
        pattern = random.choice(["random", "sorted", "nearly_sorted"])

        if pattern == "random":
            data = [random.randint(0, size * 10) for _ in range(size)]
        elif pattern == "sorted":
            data = list(range(size))
        else:
            data = _make_nearly_sorted(size)

        expected = sorted(data)
        system.run(data, task_type="sorting", expected_result=expected)

    stats = system.get_stats()
    mc_stats = stats["meta_controller"]
    print(f"History size: {mc_stats['history_size']}")
    print(f"Knowledge records: {stats['knowledge_base']['total_records']}")
    print(f"Bandit epsilon: {mc_stats['bandit']['epsilon']:.3f}")
    print(f"Last confidence: {mc_stats['last_confidence']:.0%}")

    print("\nAlgorithm Performance:")
    for algo_name, summary in stats["performance_tracker"]["summaries"].items():
        if summary and summary.get("count", 0) > 0:
            print(f"  {algo_name:15s}: avg={summary['avg_time_ms']:8.2f}ms, n={summary['count']}")


def example_4_explain_decisions():
    """Get explanations for algorithm selection with audit trail"""
    print("\n" + "=" * 70)
    print("Example 4: Decision Explanation + Audit Trail")
    print("=" * 70)

    system = AAlgoI()

    data = _make_nearly_sorted(5000)
    expected = sorted(data)
    result = system.run(data, task_type="sorting", expected_result=expected)

    explanation = system.explain_decision()
    print("\nContext:")
    for key, value in explanation["context"].items():
        print(f"  {key}: {value}")
    print(f"\nSelected: {' -> '.join(explanation['selected_algorithms'])}")
    print(f"Confidence: {explanation.get('confidence', 0):.0%}")
    print(f"Selection reason: {explanation.get('selection_reason', 'N/A')}")
    print("\nReasoning:")
    for reason in explanation["reasoning"]:
        print(f"  - {reason}")

    print("\nDecision Log:")
    log = system.get_decision_log(5)
    for entry in log:
        print(f"  [{entry['timestamp']:.0f}] {entry['chosen']:15s} | "
              f"conf={entry['confidence']:.0%} | ok={entry['success']}")


def example_5_benchmark():
    """Benchmark all algorithms"""
    print("\n" + "=" * 70)
    print("Example 5: Algorithm Benchmark")
    print("=" * 70)

    system = AAlgoI()
    data = [random.randint(0, 100000) for _ in range(10000)]

    print("\nBenchmarking all sorting algorithms...")
    results = system.benchmark(data)

    sorted_results = sorted(
        results.items(),
        key=lambda x: x[1]["time_ms"] if x[1]["success"] else float("inf")
    )

    for algo_name, result in sorted_results:
        status = "+" if result["success"] and result.get("correct", False) else "-"
        print(f"  {status} {algo_name:15s}: {result['time_ms']:8.2f}ms")

    result = system.run(data, task_type="sorting")
    chosen = system.get_stats()["active_pipeline"]
    print(f"\n* AAlgoI chose: {' -> '.join(chosen)}")


def example_6_validation_and_drift():
    """Demonstrate output validation and drift detection"""
    print("\n" + "=" * 70)
    print("Example 6: Output Validation + Drift Detection")
    print("=" * 70)

    system = AAlgoI()

    print("\nRunning 50 normal executions...")
    for i in range(50):
        data = [random.randint(0, 1000) for _ in range(100)]
        expected = sorted(data)
        system.run(data, task_type="sorting", expected_result=expected)

    print(f"Validation stats: {system.get_validation_stats()}")
    print(f"Drift stats: {system.get_drift_stats()}")


def example_7_llm_advisor():
    """LLM-powered decision explanation (requires ollama or llama.cpp)"""
    print("\n" + "=" * 70)
    print("Example 7: LLM Advisor (requires ollama/llama.cpp)")
    print("=" * 70)

    system = AAlgoI(config={
        "llm": {
            "enabled": True,
            "backend": "ollama",
            "model": "phi3:mini"
        }
    })

    data = _make_nearly_sorted(5000)
    expected = sorted(data)
    result = system.run(data, task_type="sorting", expected_result=expected)

    if system.meta_controller.llm.check_available():
        explanation = system.explain_decision()
        print(f"\nLLM Explanation: {explanation.get('llm_explanation', 'N/A')}")
    else:
        print("\nLLM not available. Start ollama: 'ollama pull phi3:mini' then 'ollama serve'")


def example_8_dag_pipeline():
    """DAG pipeline execution (parallel branches)"""
    print("\n" + "=" * 70)
    print("Example 8: DAG Pipeline (Parallel Image Processing)")
    print("=" * 70)

    try:
        import numpy as np
        system = AAlgoI(config={"enable_dag": True})

        image = np.random.rand(100, 100).astype(np.float32)

        result = system.run(image, task_type="image_processing")
        pipeline = system.get_stats()["active_pipeline"]
        print(f"\nPipeline: {' -> '.join(pipeline)}")
        print(f"DAG mode: DAG pipeline executed")

    except ImportError:
        print("\nSkipping - numpy required for image processing demo")


def _make_nearly_sorted(size):
    data = list(range(size))
    swaps = size // 20
    for _ in range(swaps):
        i, j = random.randint(0, size-1), random.randint(0, size-1)
        data[i], data[j] = data[j], data[i]
    return data


if __name__ == "__main__":
    example_1_basic_sorting()
    example_2_strategy_comparison()
    example_3_learning_over_time()
    example_4_explain_decisions()
    example_5_benchmark()
    example_6_validation_and_drift()

    print("\n" + "=" * 70)
    print("All examples completed!")
    print("=" * 70)



from core.rl.environment import AAlgoIEnv
from core.rl.replay_buffer import ReplayBuffer, EpisodeBuffer, PrioritizedReplayBuffer
from core.rl.reward_shaper import RewardShaper, AdaptiveRewardShaper, CurriculumRewardShaper
from core.rl.agents.selection_agent import PPOAgent, ActorCriticNetwork

from core.rl.powerhouse_agent import (
    HighLevelController,
    LowLevelController,
    HierarchicalAgent,
    CuriosityModule,
    WorldModel,
    MetaLearner,
    MultiTaskAgent,
)

__all__ = [
    "AAlgoIEnv",
    "ReplayBuffer", "EpisodeBuffer", "PrioritizedReplayBuffer",
    "RewardShaper", "AdaptiveRewardShaper", "CurriculumRewardShaper",
    "PPOAgent", "ActorCriticNetwork",
    "HighLevelController", "LowLevelController", "HierarchicalAgent",
    "CuriosityModule", "WorldModel", "MetaLearner", "MultiTaskAgent",
]



#!/usr/bin/env python3
"""
Verify that the package builds and installs correctly.
"""

import subprocess
import sys


def run_command(cmd, description):
    """Run a shell command and return success status."""
    print(f"  {description}...")
    try:
        result = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
        print(f"  OK - {description}")
        return True
    except subprocess.CalledProcessError as e:
        print(f"  FAILED - {description}")
        print(e.stderr)
        return False


def main():
    print("=" * 60)
    print("AAlgoI Build Verification")
    print("=" * 60)

    steps = [
        ("python -m pip install --upgrade pip", "Upgrade pip"),
        ("python -m pip install -r requirements.txt", "Install dependencies"),
        ("python -m pip install build", "Install build tool"),
        ("python -m build", "Build package"),
        ("python -m pip install -e .", "Install in editable mode"),
        ("python -c \"from pipeline import UniversalSolver; s=UniversalSolver(); print('Import OK')\"", "Test import"),
        ("python demo.py", "Run demo"),
        ("python -m pytest tests/test_full_integration.py -v", "Run test suite"),
    ]

    results = []
    for cmd, desc in steps:
        results.append(run_command(cmd, desc))

    print("=" * 60)
    if all(results):
        print("All checks passed! Package is ready to ship.")
        sys.exit(0)
    else:
        print("Some checks failed. Please review errors above.")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
AAlgoI — Artificial Algorithm Intelligence
===========================================

Self-adaptive algorithm selection & universal problem solving.

Quick start:
    >>> from aalgoi import solve
    >>> result = solve("sort these numbers", [3, 1, 4, 1, 5])
    >>> result["result"]
    [1, 1, 3, 4, 5]

    >>> from aalgoi import UniversalSolver
    >>> solver = UniversalSolver()
    >>> result = solver.solve(spec, data)

    >>> from aalgoi import SmartSolver
    >>> solver = SmartSolver()
    >>> result = solver.ask("Find shortest path", graph)
"""

from pipeline import UniversalSolver, Result
from core.smart_solver import SmartSolver
from core import (
    ProblemSpec,
    ProblemType,
    Objective,
    Constraint,
    Explainer,
    Explanation,
    AlgorithmMarketplace,
    AlgorithmMetadata,
    Benchmarker,
    DynamicRegistry,
)
from core.rl.reward_shaper import RewardShaper

__version__ = "1.1.0"

__all__ = [
    "UniversalSolver",
    "SmartSolver",
    "Result",
    "ProblemSpec",
    "ProblemType",
    "Objective",
    "Constraint",
    "solve",
    "solve_spec",
    "explain",
    "benchmark",
    "Explainer",
    "Explanation",
    "AlgorithmMarketplace",
    "AlgorithmMetadata",
    "Benchmarker",
    "DynamicRegistry",
    "RewardShaper",
    "__version__",
]

_VERSIONS = {"aalgoi": __version__}


def solve(question, data=None, config=None, **kwargs):
    """
    Parse a plain-English question and solve it in one step.

    Parameters
    ----------
    question : str
        Natural language question (e.g. "Sort these numbers fastest possible")
    data : any, optional
        Input data (list, dict, array, graph, etc.)
    config : dict, optional
        Configuration dictionary passed to UniversalSolver
    **kwargs
        Additional keyword arguments forwarded to SmartSolver.ask()

    Returns
    -------
    dict
        Result with keys: success, result, algorithm, time_ms, answer

    Example
    -------
    >>> from aalgoi import solve
    >>> r = solve("sort ascending", [3, 1, 2])
    >>> r["result"]
    [1, 2, 3]
    """
    cfg = {"rl_deterministic": False}
    if config:
        cfg.update(config)
    solver = SmartSolver(config=cfg)
    return solver.ask(question, data, **kwargs)


def solve_spec(spec, data=None, config=None):
    """
    Solve using a pre-built ProblemSpec object.

    Parameters
    ----------
    spec : ProblemSpec
        Structured problem specification
    data : any, optional
        Input data
    config : dict, optional
        Configuration dictionary

    Returns
    -------
    dict
        Result with keys: success, result, algorithm, time_ms
    """
    cfg = {"rl_deterministic": False}
    if config:
        cfg.update(config)
    solver = UniversalSolver(config=cfg)
    return solver.solve(spec, data)


def explain(result_or_algo, data=None):
    """
    Get a human-readable explanation for an algorithm choice.

    Parameters
    ----------
    result_or_algo : dict or str
        A result dict (from solve()) containing 'algorithm' key,
        or an algorithm name string
    data : any, optional
        Input data or detail string passed through to Explainer

    Returns
    -------
    Explanation
        Structured explanation object
    """
    exp = Explainer()
    if isinstance(result_or_algo, dict):
        algo_name = result_or_algo.get("algorithm", "unknown")
        return exp.explain(algo_name, detail=str(data) if data else None)
    return exp.explain(result_or_algo, detail=str(data) if data else None)


def benchmark(spec, data):
    """
    Run comparison benchmark: AAlgoI vs standard library.

    Parameters
    ----------
    spec : ProblemSpec
        Structured problem specification
    data : any
        Input data to benchmark on

    Returns
    -------
    dict
        Benchmark results with timing, speedup, and winner
    """
    bm = Benchmarker(solver=UniversalSolver())
    return bm.compare(spec, data)




import json
import time
import logging
import numpy as np
from typing import Dict, List, Any, Optional, Tuple
from collections import defaultdict

logger = logging.getLogger(__name__)


class VectorKnowledgeBase:
    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}
        self.collection_name = self.config.get("collection_name", "aalgoi_memory")
        self.max_size = self.config.get("max_size", 100000)
        self.records: List[Dict] = []
        self._index: Dict[str, List[int]] = defaultdict(list)
        self._chroma_client = None
        self._chroma_collection = None
        self._use_chromadb = False
        self._init_chromadb()

    def _init_chromadb(self):
        try:
            import chromadb
            self._chroma_client = chromadb.Client()
            try:
                self._chroma_collection = self._chroma_client.get_collection(self.collection_name)
            except ValueError:
                self._chroma_collection = self._chroma_client.create_collection(self.collection_name)
            self._use_chromadb = True
        except (ImportError, Exception):
            pass

    def store(self, context: Dict, algorithm_names: List[str], metrics: Dict):
        record = {
            "timestamp": time.time(),
            "context": {
                "features": context.get("features", {}),
                "data_profile": {
                    "type": context.get("data_profile", {}).get("type"),
                    "size": context.get("data_profile", {}).get("size"),
                    "patterns": context.get("data_profile", {}).get("patterns", {})
                },
                "constraints": context.get("constraints", {})
            },
            "algorithms": algorithm_names,
            "metrics": {
                "wall_time_ms": metrics.get("wall_time_ms"),
                "quality_score": metrics.get("quality_score"),
                "success": metrics.get("success"),
                "within_budget": metrics.get("within_budget"),
                "score": metrics.get("score", 0)
            }
        }

        self.records.append(record)
        idx = len(self.records) - 1
        for algo in algorithm_names:
            self._index[algo].append(idx)

        if self._use_chromadb and self._chroma_collection is not None:
            try:
                doc_id = f"rec_{idx}"
                doc_text = json.dumps({
                    "type": context.get("data_profile", {}).get("type"),
                    "size": context.get("data_profile", {}).get("size"),
                    "is_sorted": context.get("data_profile", {}).get("patterns", {}).get("is_sorted"),
                    "is_nearly_sorted": context.get("data_profile", {}).get("patterns", {}).get("is_nearly_sorted"),
                    "priority": context.get("constraints", {}).get("priority"),
                    "algorithms": algorithm_names
                })
                metadata = {
                    "algorithms": ",".join(algorithm_names),
                    "score": str(metrics.get("score", 0)),
                    "success": str(metrics.get("success", True))
                }
                self._chroma_collection.add(
                    documents=[doc_text],
                    metadatas=[metadata],
                    ids=[doc_id]
                )
            except Exception:
                pass

        if len(self.records) > self.max_size:
            self._prune_old_records()

    def query_similar(self, context: Dict, top_k: int = 5) -> List[Dict]:
        if self._use_chromadb and self._chroma_collection is not None:
            try:
                query_text = json.dumps({
                    "type": context.get("data_profile", {}).get("type"),
                    "size": context.get("data_profile", {}).get("size"),
                    "is_sorted": context.get("data_profile", {}).get("patterns", {}).get("is_sorted"),
                    "is_nearly_sorted": context.get("data_profile", {}).get("patterns", {}).get("is_nearly_sorted"),
                    "priority": context.get("constraints", {}).get("priority")
                })
                results = self._chroma_collection.query(
                    query_texts=[query_text],
                    n_results=min(top_k, len(self.records))
                )
                if results and results.get("metadatas") and results["metadatas"][0]:
                    return self._chroma_to_records(results)
            except Exception:
                pass

        return self._fallback_similarity_search(context, top_k)

    def _chroma_to_records(self, results: Dict) -> List[Dict]:
        records = []
        for i, meta in enumerate(results["metadatas"][0]):
            rec = {
                "algorithms": meta.get("algorithms", "").split(",") if meta.get("algorithms") else [],
                "metrics": {
                    "score": float(meta.get("score", 0)),
                    "success": meta.get("success", "True") == "True"
                }
            }
            if results.get("documents") and results["documents"][0] and i < len(results["documents"][0]):
                try:
                    doc = json.loads(results["documents"][0][i])
                    rec["context"] = {"data_profile": doc}
                except (json.JSONDecodeError, IndexError):
                    pass
            records.append(rec)
        return records

    def _fallback_similarity_search(self, context: Dict, top_k: int) -> List[Dict]:
        if not self.records:
            return []

        current_features = context.get("features", {})
        current_vec = self._features_to_vector(current_features)

        similarities = []
        for record in self.records:
            record_features = record.get("context", {}).get("features", {})
            record_vec = self._features_to_vector(record_features)
            similarity = self._cosine_similarity(current_vec, record_vec)
            similarities.append((similarity, record))

        similarities.sort(key=lambda x: x[0], reverse=True)
        return [record for _, record in similarities[:top_k]]

    def penalize(self, algorithm_name: str, context: Dict):
        penalty_record = {
            "timestamp": time.time(),
            "context": {"features": context.get("features", {})},
            "algorithms": [algorithm_name],
            "metrics": {
                "score": -1.0,
                "success": False,
                "penalty": True
            }
        }
        self.records.append(penalty_record)
        self._index[algorithm_name].append(len(self.records) - 1)

    def discount_old_records(self, factor: float = 0.5):
        cutoff = time.time() - 86400 * 7
        for record in self.records:
            if record.get("timestamp", 0) < cutoff:
                metrics = record.get("metrics", {})
                score = metrics.get("score", 0)
                if score > 0:
                    metrics["score"] = score * factor

    def get_algorithm_stats(self, algorithm_name: str) -> Dict[str, Any]:
        indices = self._index.get(algorithm_name, [])
        if not indices:
            return {}

        records = [self.records[i] for i in indices]
        times = [r["metrics"].get("wall_time_ms", 0) for r in records if r["metrics"].get("wall_time_ms") is not None]
        qualities = [r["metrics"].get("quality_score", 0) for r in records if r["metrics"].get("quality_score") is not None]
        scores = [r["metrics"].get("score", 0) for r in records if r["metrics"].get("score") is not None]

        return {
            "total_executions": len(records),
            "avg_time_ms": np.mean(times) if times else 0,
            "avg_quality": np.mean(qualities) if qualities else 0,
            "avg_score": np.mean(scores) if scores else 0,
            "success_rate": sum(1 for r in records if r["metrics"].get("success")) / len(records) if records else 0,
            "last_used": max(r.get("timestamp", 0) for r in records) if records else None
        }

    def get_all_stats(self) -> Dict[str, Dict]:
        return {
            algo: self.get_algorithm_stats(algo)
            for algo in self._index.keys()
        }

    def get_best_algorithm(self, context: Dict, metric: str = "quality_score") -> Optional[str]:
        similar = self.query_similar(context, top_k=20)
        if not similar:
            return None

        unique_algos = set()
        for record in similar:
            for algo in record.get("algorithms", []):
                if algo:
                    unique_algos.add(algo)

        if len(unique_algos) == 1:
            return list(unique_algos)[0]

        algo_scores = defaultdict(list)
        for record in similar:
            for algo in record.get("algorithms", []):
                score = record.get("metrics", {}).get(metric)
                if score is not None:
                    algo_scores[algo].append(score)

        if not algo_scores:
            return None

        best_algo = max(algo_scores.keys(), key=lambda a: np.mean(algo_scores[a]))
        return best_algo

    def _features_to_vector(self, features: Dict) -> np.ndarray:
        feature_names = [
            "data_size_log", "is_numeric", "is_sorted", "is_nearly_sorted",
            "cpu_free", "mem_free_ratio", "cpu_count", "time_budget_norm",
            "priority_speed", "priority_accuracy"
        ]
        return np.array([features.get(f, 0.0) for f in feature_names])

    def _cosine_similarity(self, a: np.ndarray, b: np.ndarray) -> float:
        norm_a = np.linalg.norm(a)
        norm_b = np.linalg.norm(b)
        if norm_a == 0 or norm_b == 0:
            return 0.0
        return np.dot(a, b) / (norm_a * norm_b)

    def _prune_old_records(self):
        keep_count = self.max_size // 2
        self.records = self.records[-keep_count:]
        self._index = defaultdict(list)
        for i, record in enumerate(self.records):
            for algo in record.get("algorithms", []):
                self._index[algo].append(i)


    def merge(self, global_knowledge: Dict):
        """
        Merge knowledge from the federated network.
        Integrate remote algorithm performance metrics.
        """
        if not global_knowledge:
            return

        import time

        for algo_name, metrics in global_knowledge.items():
            entry = {
                "algorithm": algo_name,
                "score": metrics.get("avg_reward", 0),
                "success_rate": metrics.get("success_rate", 0),
                "context_patterns": metrics.get("context_patterns", []),
                "source": "federated",
                "timestamp": time.time(),
            }
            self.records.append(entry)
            idx = len(self.records) - 1
            self._index[algo_name].append(idx)

        logger.info("Merged %d federated knowledge entries", len(global_knowledge))

    def get_top_performing(self, n: int = 10) -> Dict:
        """Extract top N best performing algorithms for sharing."""
        sorted_records = sorted(
            self.records,
            key=lambda x: x.get("metrics", {}).get("score", x.get("score", 0)),
            reverse=True,
        )[:n]

        return {
            r.get("algorithm", "unknown"): {
                "avg_reward": r.get("metrics", {}).get("score", r.get("score", 0)),
                "success_rate": 1.0
                if r.get("metrics", {}).get("score", 0) > 5
                else 0.5,
                "context_patterns": [],
            }
            for r in sorted_records
            if r.get("algorithm")
        }


KnowledgeBase = VectorKnowledgeBase



import random
import math
from algorithms.base import Algorithm


class GreedyKnapsack(Algorithm):
    def __init__(self):
        super().__init__()
        self.name = "greedy_knapsack"
        self.time_complexity = "O(n log n)"
        self.space_complexity = "O(1)"
        self.tags = ["optimization", "knapsack", "greedy", "approximation"]
        self.best_for = ["resource_allocation", "fast_approximation"]

    def validate_output(self, input_data, output_data):
        return isinstance(output_data, dict) and "selected" in output_data and "value" in output_data

    def process(self, data):
        items = data['items']
        capacity = data['capacity']

        indexed_items = sorted(
            enumerate(items), 
            key=lambda x: x[1]['value']/x[1]['weight'], 
            reverse=True
        )

        total_weight = 0
        total_value = 0
        selected_indices = []

        for idx, item in indexed_items:
            if total_weight + item['weight'] <= capacity:
                selected_indices.append(idx)
                total_weight += item['weight']
                total_value += item['value']

        return {
            'selected': selected_indices,
            'value': total_value,
            'weight': total_weight
        }


class SimulatedAnnealing(Algorithm):
    def __init__(self):
        super().__init__()
        self.name = "simulated_annealing"
        self.time_complexity = "O(iterations)"
        self.space_complexity = "O(n)"
        self.tags = ["optimization", "metaheuristic", "combinatorial"]
        self.best_for = ["tsp", "knapsack", "complex_landscape"]

    def validate_output(self, input_data, output_data):
        return isinstance(output_data, dict) and "selected" in output_data and "value" in output_data

    def process(self, data):
        items = data['items']
        capacity = data['capacity']
        iterations = data.get('iterations', 1000)

        def evaluate(state):
            val = sum(items[i]['value'] for i in range(len(items)) if state[i])
            wt = sum(items[i]['weight'] for i in range(len(items)) if state[i])
            return val if wt <= capacity else 0

        state = [random.choice([0, 1]) for _ in range(len(items))]
        best_state = state[:]
        best_val = evaluate(state)
        temp = 1000.0

        for i in range(iterations):
            neighbor = state[:]
            idx = random.randint(0, len(items)-1)
            neighbor[idx] = 1 - neighbor[idx]

            n_val = evaluate(neighbor)
            current_val = evaluate(state)

            if n_val > current_val or random.random() < math.exp((n_val - current_val) / temp):
                state = neighbor
                if n_val > best_val:
                    best_val = n_val
                    best_state = neighbor

            temp *= 0.995

        final_wt = sum(items[i]['weight'] for i in range(len(items)) if best_state[i])
        return {
            'selected': [i for i, x in enumerate(best_state) if x],
            'value': best_val,
            'weight': final_wt
        }



import os
import sys

import click
import numpy as np

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from core.smart_solver import SmartSolver


@click.group()
def ml():
    """Machine Learning tasks via aalgoi CLI."""
    pass


@ml.command("train-word2vec")
@click.option("--corpus", type=click.Path(exists=True), required=True,
              help="Path to text corpus file (one sentence per line)")
@click.option("--domain", default="general",
              help="Domain: legal, medical, finance, etc.")
@click.option("--vector-size", default=100, type=int,
              help="Embedding dimension")
@click.option("--epochs", default=10, type=int, help="Training epochs")
@click.option("--output", type=click.Path(), help="Save model to file")
def train_word2vec(corpus, domain, vector_size, epochs, output):
    """Train Word2Vec embeddings on domain-specific corpus."""
    with open(corpus, "r", encoding="utf-8") as f:
        sentences = [line.strip() for line in f if line.strip()]

    if not sentences:
        click.echo("Error: Empty corpus file", err=True)
        sys.exit(1)

    solver = SmartSolver()
    data = {
        "corpus": sentences,
        "vector_size": vector_size,
        "epochs": epochs,
        "domain": domain,
    }
    question = f"Train Word2Vec on {domain} corpus with {vector_size} dimensions"
    result = solver.ask(question, data)

    if result.get("success"):
        model = result.get("result", {}).get("model")
        vocab_size = result.get("result", {}).get("vocabulary_size", 0)

        click.echo(click.style("SUCCESS: Trained Word2Vec", fg="green"))
        click.echo(f"  Vocabulary size: {vocab_size}")
        click.echo(f"  Vector size: {vector_size}")
        click.echo(f"  Domain: {domain}")

        if output and model:
            model.save(output)
            click.echo(f"  Model saved to: {output}")
    else:
        click.echo(
            click.style(f"FAILED: {result.get('error', 'Unknown error')}", fg="red")
        )
        sys.exit(1)


@ml.command("similar-words")
@click.option("--model", type=click.Path(exists=True), required=True,
              help="Path to trained Word2Vec model")
@click.option("--word", required=True, help="Input word")
@click.option("--topn", default=5, type=int, help="Number of similar words")
@click.option("--domain", default=None, help="Domain bias for results")
def similar_words(model, word, topn, domain):
    """Find semantically similar words using trained embeddings."""
    try:
        from gensim.models import Word2Vec as W2V
        w2v_model = W2V.load(model)
    except Exception as e:
        click.echo(f"Error loading model: {e}", err=True)
        sys.exit(1)

    solver = SmartSolver()
    data = {
        "model": w2v_model,
        "input_word": word,
        "topn": topn,
        "domain": domain or "general",
    }
    result = solver.ask(f"Find {topn} words similar to {word}", data)

    if result.get("success"):
        similar = result.get("result", {}).get("similar_words", [])

        if not similar:
            click.echo(f"'{word}' not found in vocabulary")
            sys.exit(1)

        click.echo(f"Words similar to '{word}':")
        for item in similar:
            click.echo(f"  {item['word']}: {item['similarity']:.3f}")
    else:
        click.echo(
            click.style(f"FAILED: {result.get('error', 'Unknown error')}", fg="red")
        )
        sys.exit(1)


@ml.command("visualize-embeddings")
@click.option("--model", type=click.Path(exists=True), required=True,
              help="Path to trained Word2Vec model")
@click.option("--words", multiple=True, required=True,
              help="Words to visualize (provide multiple --words flags)")
@click.option("--method", type=click.Choice(["pca", "tsne"]), default="pca",
              help="Dimensionality reduction method")
@click.option("--output", type=click.Path(), help="Save visualization image to file")
def visualize_embeddings(model, words, method, output):
    """Visualize word embeddings in 2D."""
    try:
        from gensim.models import Word2Vec as W2V
        w2v_model = W2V.load(model)
    except Exception as e:
        click.echo(f"Error loading model: {e}", err=True)
        sys.exit(1)

    embeddings = []
    valid_words = []
    for word in words:
        if word in w2v_model.wv:
            embeddings.append(w2v_model.wv[word])
            valid_words.append(word)
        else:
            click.echo(f"Warning: '{word}' not in vocabulary, skipping")

    if not embeddings:
        click.echo(click.style("FAILED: None of the specified words are in the vocabulary", fg="red"))
        sys.exit(1)

    solver = SmartSolver()
    data = {
        "embeddings": np.array(embeddings),
        "words": valid_words,
        "method": method,
    }
    result = solver.ask(f"Visualize {len(valid_words)} words using {method}", data)

    if result.get("success"):
        viz_data = result.get("result", {}).get("visualization_data", {})

        click.echo(click.style(f"SUCCESS: {method.upper()} visualization generated", fg="green"))
        click.echo(f"  Words: {valid_words}")
        click.echo(f"  Explained variance: {viz_data.get('explained_variance', 'N/A')}")

        if output:
            try:
                import matplotlib.pyplot as plt
                coords = np.array(viz_data["coordinates"])
                plt.figure(figsize=(10, 8))
                plt.scatter(coords[:, 0], coords[:, 1])
                for i, word in enumerate(valid_words):
                    plt.annotate(word, (coords[i, 0], coords[i, 1]))
                plt.title(f"Word Embeddings ({method.upper()})")
                plt.savefig(output)
                click.echo(f"  Saved to: {output}")
            except ImportError:
                click.echo("Warning: matplotlib not available, cannot save plot")
    else:
        click.echo(
            click.style(f"FAILED: {result.get('error', 'Unknown error')}", fg="red")
        )
        sys.exit(1)




import json
import subprocess
import time
from typing import Any, Dict, List, Optional, Tuple


SELECTION_PROMPT = """You are an algorithm selection expert. Given this context:

Data: size={size}, type={data_type}, sorted={is_sorted}, nearly_sorted={is_nearly_sorted}
System: cpu_free={cpu_free:.0%}, mem_free={mem_free:.0%}, cpu_count={cpu_count}
Constraints: time_budget={time_budget}ms, priority={priority}

Available algorithms:
{algorithms}

Select the optimal algorithm. Return ONLY valid JSON:
{{"algorithm": "name", "reason": "brief reason"}}
"""

EXPLANATION_PROMPT = """Explain why the algorithm "{chosen}" was selected for this context:

Data: size={size}, type={data_type}, sorted={is_sorted}, nearly_sorted={is_nearly_sorted}
System: cpu_free={cpu_free:.0%}, mem_free={mem_free:.0%}
Priority: {priority}

Result: time={time_ms:.1f}ms, quality={quality:.2f}, success={success}

Return ONLY valid JSON:
{{"explanation": "your explanation", "confidence": 0.0-1.0, "suggestion": "optional improvement"}}
"""

PARAM_PROMPT = """Suggest optimal parameters for algorithm "{algo_name}" given:

Data: size={size}, type={data_type}
System: cpu_free={cpu_free:.0%}, mem_free={mem_free:.0%}
Priority: {priority}

Known parameters: {known_params}

Return ONLY valid JSON with parameter names and values.
"""


class LLMAdapter:
    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}
        self.backend = self.config.get("backend", "ollama")
        self.model = self.config.get("model", "phi3:mini")
        self.ollama_url = self.config.get("ollama_url", "http://localhost:11434")
        self.llamacpp_url = self.config.get("llamacpp_url", "http://localhost:8080")
        self.max_tokens = self.config.get("max_tokens", 256)
        self.temperature = self.config.get("temperature", 0.1)
        self.enabled = self.config.get("enabled", False)
        self._available = None

    def check_available(self) -> bool:
        if self._available is not None:
            return self._available

        if not self.enabled:
            self._available = False
            return False

        try:
            import requests
            if self.backend == "ollama":
                r = requests.get(f"{self.ollama_url}/api/tags", timeout=2)
                self._available = r.status_code == 200
            else:
                r = requests.get(f"{self.llamacpp_url}/health", timeout=2)
                self._available = r.status_code == 200
        except Exception:
            self._available = False

        return self._available

    def analyze_context(self, context: Dict) -> Dict[str, Any]:
        if not self.enabled:
            return {"algorithm": None, "reason": "LLM disabled"}

        if not self.check_available():
            return {"algorithm": None, "reason": "LLM not available"}

        fp = context.get("features", {})
        dp = context.get("data_profile", {})
        env = context.get("environment", {})
        cons = context.get("constraints", {})

        cpu = fp.get("cpu_free", 0.5)
        mem = fp.get("mem_free_ratio", 0.5)
        cpu_count = fp.get("cpu_count", 1)

        task_type = context.get("task_type", "sorting")
        algo_descriptions = self._get_algo_descriptions(task_type)

        prompt = SELECTION_PROMPT.format(
            size=dp.get("size", "unknown"),
            data_type=dp.get("type", "unknown"),
            is_sorted=dp.get("patterns", {}).get("is_sorted", False),
            is_nearly_sorted=dp.get("patterns", {}).get("is_nearly_sorted", False),
            cpu_free=cpu,
            mem_free=mem,
            cpu_count=cpu_count,
            time_budget=cons.get("time_budget_ms", 500),
            priority=cons.get("priority", "balanced"),
            algorithms=algo_descriptions
        )

        response = self._query(prompt)
        return self._parse_selection(response)

    def explain_decision(self, context: Dict, chosen: str, result_metrics: Dict) -> str:
        if not self.enabled or not self.check_available():
            return "LLM explanation unavailable"

        fp = context.get("features", {})
        dp = context.get("data_profile", {})

        prompt = EXPLANATION_PROMPT.format(
            chosen=chosen,
            size=dp.get("size", "unknown"),
            data_type=dp.get("type", "unknown"),
            is_sorted=dp.get("patterns", {}).get("is_sorted", False),
            is_nearly_sorted=dp.get("patterns", {}).get("is_nearly_sorted", False),
            cpu_free=fp.get("cpu_free", 0.5),
            mem_free=fp.get("mem_free_ratio", 0.5),
            priority=context.get("constraints", {}).get("priority", "balanced"),
            time_ms=result_metrics.get("wall_time_ms", 0),
            quality=result_metrics.get("quality_score", 0),
            success=result_metrics.get("success", False)
        )

        response = self._query(prompt)
        try:
            return json.loads(response).get("explanation", response)
        except (json.JSONDecodeError, ValueError):
            return response

    def suggest_parameters(self, context: Dict, algo_name: str) -> Dict:
        if not self.enabled or not self.check_available():
            return {}

        fp = context.get("features", {})
        dp = context.get("data_profile", {})

        known = self._get_known_params(algo_name)

        prompt = PARAM_PROMPT.format(
            algo_name=algo_name,
            size=dp.get("size", "unknown"),
            data_type=dp.get("type", "unknown"),
            cpu_free=fp.get("cpu_free", 0.5),
            mem_free=fp.get("mem_free_ratio", 0.5),
            priority=context.get("constraints", {}).get("priority", "balanced"),
            known_params=json.dumps(known)
        )

        response = self._query(prompt)
        return self._try_parse_json(response)

    def _query(self, prompt: str) -> str:
        try:
            import requests
            if self.backend == "ollama":
                return self._query_ollama(prompt)
            else:
                return self._query_llamacpp(prompt)
        except (ImportError, Exception) as e:
            return json.dumps({"error": str(e)})

    def _query_ollama(self, prompt: str) -> str:
        import requests
        payload = {
            "model": self.model,
            "prompt": prompt,
            "stream": False,
            "options": {
                "num_predict": self.max_tokens,
                "temperature": self.temperature
            }
        }
        r = requests.post(f"{self.ollama_url}/api/generate", json=payload, timeout=30)
        if r.status_code == 200:
            return r.json().get("response", "")
        return json.dumps({"error": f"Ollama error: {r.status_code}"})

    def _query_llamacpp(self, prompt: str) -> str:
        import requests
        payload = {
            "prompt": prompt,
            "n_predict": self.max_tokens,
            "temperature": self.temperature
        }
        r = requests.post(f"{self.llamacpp_url}/completion", json=payload, timeout=30)
        if r.status_code == 200:
            return r.json().get("content", "")
        return json.dumps({"error": f"llama.cpp error: {r.status_code}"})

    def _parse_selection(self, response: str) -> Dict:
        try:
            parsed = json.loads(response)
            if isinstance(parsed, dict) and "algorithm" in parsed:
                return parsed
        except (json.JSONDecodeError, ValueError):
            pass

        for line in response.strip().split('\n'):
            line = line.strip()
            if line.startswith('{') and line.endswith('}'):
                try:
                    parsed = json.loads(line)
                    if isinstance(parsed, dict) and "algorithm" in parsed:
                        return parsed
                except json.JSONDecodeError:
                    continue

        return {"algorithm": None, "reason": response[:200]}

    def _try_parse_json(self, text: str) -> Dict:
        try:
            return json.loads(text)
        except (json.JSONDecodeError, ValueError):
            for line in text.strip().split('\n'):
                line = line.strip()
                if line.startswith('{') and line.endswith('}'):
                    try:
                        return json.loads(line)
                    except json.JSONDecodeError:
                        continue
        return {"raw": text[:200]}

    def _get_algo_descriptions(self, task_type: str) -> str:
        descriptions = {
            "sorting": [
                "quicksort - O(n log n) avg, in-place, fast random data, bad nearly-sorted",
                "insertion_sort - O(n²) worst, O(n) best, great tiny/nearly-sorted data",
                "merge_sort - O(n log n) guaranteed, stable, uses O(n) memory",
                "timsort - O(n log n), hybrid, great real-world/nearly-sorted, Python built-in",
                "radix_sort - O(nk) linear for integers, needs small digit range",
                "heap_sort - O(n log n) in-place, O(1) space, good memory-constrained"
            ],
            "image_processing": [
                "gaussian_blur - O(n), fast noise reduction, blurs edges",
                "median_filter - O(n), great salt-pepper noise, preserves edges",
                "bilateral_filter - O(n), edge-preserving denoise, slower",
                "sobel_edge - O(n), fast edge detection",
                "clahe - O(n), adaptive contrast enhancement, good low-contrast images"
            ],
            "ml": [
                "kmeans - O(n*k*i), simple clustering, needs k specified",
                "dbscan - O(n log n), density clustering, detects outliers",
                "random_forest - ensemble classifier, high accuracy, slower",
                "linear_regression - O(n*d²), fast regression, interpretable"
            ]
        }
        algos = descriptions.get(task_type, descriptions["sorting"])
        return "\n".join(f"  - {a}" for a in algos)

    def _get_known_params(self, algo_name: str) -> Dict:
        known = {
            "quicksort": {"in_place": False},
            "timsort": {},
            "merge_sort": {},
            "insertion_sort": {},
            "heap_sort": {},
            "radix_sort": {},
            "gaussian_blur": {"sigma": 1.0},
            "median_filter": {"filter_size": 3},
            "bilateral_filter": {"sigma_color": 0.1, "sigma_spatial": 1.0},
            "sobel_edge": {},
            "clahe": {"clip_limit": 0.03},
            "kmeans": {"n_clusters": 3, "max_iter": 100},
            "dbscan": {"eps": 0.5, "min_samples": 5},
            "random_forest": {"n_estimators": 100, "max_depth": 10},
            "linear_regression": {"fit_intercept": True}
        }
        return known.get(algo_name, {})

    def get_stats(self) -> Dict:
        return {
            "enabled": self.enabled,
            "available": self._available,
            "backend": self.backend,
            "model": self.model
        }



#!/usr/bin/env python3
"""Pre-train the RL algorithm selector on synthetic data across all 3 domains."""

import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
)

from training.pretrain import RLPretrainer


def main():
    config = {
        "episodes_per_level": 800,
        "batch_size": 128,
        "learning_rate": 3e-4,
        "save_dir": "checkpoints/pretrain_v2",
        "time_budget_ms": 500.0,
    }

    trainer = RLPretrainer(save_dir=config["save_dir"], config=config)
    summary = trainer.run_training(total_episodes=2000)

    print("\n=== Training Summary ===")
    for k, v in summary.items():
        print(f"  {k}: {v}")


if __name__ == "__main__":
    main()



from interface.cli import main as cli_main
from interface.api import create_app, run
from interface.web_ui import create_interface, launch
from interface.nl_parser import parse_description, parse_solve_input, extract_data_from_description

__all__ = [
    "cli_main",
    "create_app", "run",
    "create_interface", "launch",
    "parse_description", "parse_solve_input", "extract_data_from_description",
]



import numpy as np
import random
from collections import deque
from typing import Dict, Any, Tuple, Optional


class ReplayBuffer:
    def __init__(self, capacity: int = 10000):
        self.buffer = deque(maxlen=capacity)
        self.capacity = capacity

    def push(self, state: np.ndarray, action: int, reward: float,
             next_state: np.ndarray, done: bool, info: Dict = None):
        self.buffer.append({
            "state": state, "action": action, "reward": reward,
            "next_state": next_state, "done": done, "info": info or {}
        })

    def sample(self, batch_size: int) -> Dict[str, np.ndarray]:
        batch = random.sample(self.buffer, min(batch_size, len(self.buffer)))
        return {
            "states": np.array([t["state"] for t in batch]),
            "actions": np.array([t["action"] for t in batch]),
            "rewards": np.array([t["reward"] for t in batch]),
            "next_states": np.array([t["next_state"] for t in batch]),
            "dones": np.array([t["done"] for t in batch]),
        }

    def __len__(self) -> int:
        return len(self.buffer)

    def clear(self):
        self.buffer.clear()


class EpisodeBuffer:
    def __init__(self):
        self.states: list = []
        self.actions: list = []
        self.rewards: list = []
        self.dones: list = []
        self.log_probs: list = []
        self.values: list = []

    def push(self, state: np.ndarray, action: int, reward: float,
             done: bool, log_prob: float = None, value: float = None):
        self.states.append(state)
        self.actions.append(action)
        self.rewards.append(reward)
        self.dones.append(done)
        if log_prob is not None:
            self.log_probs.append(log_prob)
        if value is not None:
            self.values.append(value)

    def get(self) -> Dict[str, np.ndarray]:
        result = {
            "states": np.array(self.states),
            "actions": np.array(self.actions),
            "rewards": np.array(self.rewards),
            "dones": np.array(self.dones),
        }
        if self.log_probs:
            result["log_probs"] = np.array(self.log_probs)
        if self.values:
            result["values"] = np.array(self.values)
        return result

    def __len__(self) -> int:
        return len(self.states)

    def clear(self):
        for attr in ("states", "actions", "rewards", "dones", "log_probs", "values"):
            getattr(self, attr).clear()


class PrioritizedReplayBuffer(ReplayBuffer):
    def __init__(self, capacity: int = 10000, alpha: float = 0.6):
        super().__init__(capacity)
        self.priorities = deque(maxlen=capacity)
        self.alpha = alpha

    def push(self, state, action, reward, next_state, done, info=None, priority=None):
        super().push(state, action, reward, next_state, done, info)
        if priority is None:
            priority = max(self.priorities) if self.priorities else 1.0
        self.priorities.append(priority)

    def sample(self, batch_size: int, beta: float = 0.4) -> Tuple[Dict, np.ndarray, np.ndarray]:
        priorities = np.array(self.priorities)
        probs = priorities ** self.alpha
        probs /= probs.sum()
        indices = np.random.choice(len(self.buffer), batch_size, p=probs, replace=False)
        batch = {
            k: np.array([self.buffer[i][k] for i in indices])
            for k in ("state", "action", "reward", "next_state", "done")
        }
        weights = (len(self.buffer) * probs[indices]) ** (-beta)
        weights /= weights.max()
        return batch, indices, weights

    def update_priorities(self, indices: np.ndarray, priorities: np.ndarray):
        for idx, priority in zip(indices, priorities):
            self.priorities[idx] = priority + 1e-6




import time
import traceback
from typing import Any, Dict, List, Optional, Tuple, Callable
import numpy as np

class PerformanceTracker:
    """
    Advanced performance tracking with multi-dimensional metrics.
    Tracks time, memory, accuracy, and custom quality scores.
    """

    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}
        self.measurements: List[Dict] = []
        self._baseline_measurements: Dict[str, List[float]] = {}

    def evaluate(self, fn: Callable, data: Any, context: Dict, 
                 expected_result: Any = None) -> Tuple[Any, Dict]:
        """
        Execute function and measure comprehensive performance metrics.
        """
        import psutil
        import os

        process = psutil.Process(os.getpid())

        # Pre-execution metrics
        mem_before = process.memory_info().rss / (1024 * 1024)  # MB
        cpu_before = process.cpu_percent()

        # Execute with timing
        start_time = time.perf_counter()
        start_cpu = time.process_time()

        try:
            result = fn(data)
            success = True
            error = None
        except Exception as e:
            result = None
            success = False
            error = str(e)
            traceback_str = traceback.format_exc()

        end_time = time.perf_counter()
        end_cpu = time.process_time()

        # Post-execution metrics
        mem_after = process.memory_info().rss / (1024 * 1024)  # MB
        cpu_after = process.cpu_percent()

        # Calculate metrics
        wall_time_ms = (end_time - start_time) * 1000
        cpu_time_ms = (end_cpu - start_cpu) * 1000
        memory_delta_mb = mem_after - mem_before

        # Quality metrics
        quality_score = self._compute_quality_score(
            result, expected_result, context, success
        )

        # Budget compliance
        time_budget = context.get("constraints", {}).get("time_budget_ms", float('inf'))
        within_budget = wall_time_ms <= time_budget

        metrics = {
            "wall_time_ms": wall_time_ms,
            "cpu_time_ms": cpu_time_ms,
            "memory_delta_mb": memory_delta_mb,
            "memory_peak_mb": max(mem_before, mem_after),
            "within_budget": within_budget,
            "budget_utilization": wall_time_ms / time_budget if time_budget != float('inf') else 0,
            "quality_score": quality_score,
            "success": success,
            "error": error,
            "throughput": len(data) / (wall_time_ms / 1000) if wall_time_ms > 0 and hasattr(data, '__len__') else 0,
        }

        # Record measurement
        measurement = {
            "timestamp": time.time(),
            "context_features": context.get("features", {}),
            "metrics": metrics,
            "algorithm_names": context.get("algorithms", []),
        }
        self.measurements.append(measurement)

        # Update baselines
        for algo_name in context.get("algorithms", []):
            if algo_name not in self._baseline_measurements:
                self._baseline_measurements[algo_name] = []
            self._baseline_measurements[algo_name].append(wall_time_ms)

        return result, metrics

    def _compute_quality_score(self, result: Any, expected: Any, 
                               context: Dict, success: bool) -> float:
        """
        Compute quality score from 0 to 1.
        """
        if not success:
            return 0.0

        if expected is not None:
            # Compare with expected result
            try:
                if isinstance(result, list) and isinstance(expected, list):
                    if len(result) != len(expected):
                        return 0.5  # Partial credit

                    # Check if sorted correctly
                    if result == sorted(expected):
                        correct_count = sum(1 for a, b in zip(result, expected) if a == b)
                        return correct_count / len(expected)
                    else:
                        return 0.0  # Not sorted
                else:
                    return 1.0 if result == expected else 0.0
            except:
                return 0.5

        # No expected result - use heuristics
        if isinstance(result, list):
            # Check if result is sorted
            is_sorted = all(result[i] <= result[i+1] for i in range(len(result)-1))
            return 1.0 if is_sorted else 0.0

        return 1.0 if result is not None else 0.0

    def get_performance_summary(self, algorithm_name: Optional[str] = None) -> Dict:
        """
        Get performance summary for an algorithm or overall.
        """
        if algorithm_name:
            measurements = [
                m for m in self.measurements 
                if algorithm_name in m.get("algorithm_names", [])
            ]
        else:
            measurements = self.measurements

        if not measurements:
            return {}

        times = [m["metrics"]["wall_time_ms"] for m in measurements]
        qualities = [m["metrics"]["quality_score"] for m in measurements]

        return {
            "count": len(measurements),
            "avg_time_ms": np.mean(times),
            "std_time_ms": np.std(times),
            "min_time_ms": np.min(times),
            "max_time_ms": np.max(times),
            "avg_quality": np.mean(qualities),
            "success_rate": sum(1 for m in measurements if m["metrics"]["success"]) / len(measurements),
            "budget_compliance": sum(1 for m in measurements if m["metrics"]["within_budget"]) / len(measurements)
        }

    def get_baseline(self, algorithm_name: str) -> Optional[float]:
        """Get average baseline performance for an algorithm."""
        measurements = self._baseline_measurements.get(algorithm_name, [])
        if measurements:
            return np.median(measurements)
        return None




import ast
import inspect
import textwrap
import functools
import types
from typing import Any, Dict, List, Optional, Callable
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import copy
import time

from core.pipeline_graph import PipelineGraph


class DynamicCompositor:
    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}
        self._modification_history: List[Dict] = []
        self._cache_registry: Dict[str, Any] = {}
        self._ast_cache: Dict[str, ast.AST] = {}
        self._use_dag = self.config.get("enable_dag", True)

    def build_pipeline(self, algorithms: List[Any], context: Dict[str, Any]) -> List[Any]:
        if self._use_dag and len(algorithms) > 1:
            return self._build_dag_pipeline(algorithms, context)
        return self._build_linear_pipeline(algorithms, context)

    def _build_linear_pipeline(self, algorithms: List[Any], context: Dict[str, Any]) -> List[Any]:
        pipeline = []

        for i, algo in enumerate(algorithms):
            adapted = algo.clone() if hasattr(algo, 'clone') else copy.deepcopy(algo)
            adapted = self._adapt(adapted, context, pipeline_position=i)
            pipeline.append(adapted)

        return pipeline

    def _build_dag_pipeline(self, algorithms: List[Any], context: Dict[str, Any]) -> List[Any]:
        graph = PipelineGraph()
        features = context.get("features", {})

        cpu_free = features.get("cpu_free", 0.5)
        parallel_ok = cpu_free > 0.5 and len(algorithms) > 1

        for i, algo in enumerate(algorithms):
            adapted = algo.clone() if hasattr(algo, 'clone') else copy.deepcopy(algo)
            adapted = self._adapt(adapted, context, pipeline_position=i)

            if parallel_ok and i > 0 and self._can_parallelize(adapted, context):
                graph.add_algorithm(f"step_{i}", adapted,
                                    depends_on=[f"step_{j}" for j in range(i) if not self._can_parallelize(
                                        algorithms[j], context)])
            else:
                deps = [f"step_{i-1}"] if i > 0 else []
                graph.add_algorithm(f"step_{i}", adapted, depends_on=deps)

        return graph

    def _can_parallelize(self, algo: Any, context: Dict) -> bool:
        if any(tag in getattr(algo, 'tags', []) for tag in ["sorting", "stable"]):
            return False
        if hasattr(algo, 'name') and algo.name in ['timsort', 'merge_sort', 'heap_sort']:
            return False
        return True

    def _adapt(self, algo: Any, context: Dict, pipeline_position: int = 0) -> Any:
        features = context.get("features", {})
        predictions = context.get("predictions", {})
        constraints = context.get("constraints", {})

        algo = self._parametric_tune(algo, context)

        if self._should_cache(algo, context):
            algo = self._inject_memoization(algo)

        if self._should_parallelize(algo, context):
            algo = self._inject_parallelization(algo, context)

        if pipeline_position > 0:
            algo = self._optimize_intermediate(algo, context)

        if self._should_ast_optimize(algo, context):
            algo = self._ast_optimize(algo, context)

        return algo

    def _parametric_tune(self, algo: Any, context: Dict) -> Any:
        features = context.get("features", {})
        constraints = context.get("constraints", {})

        params = {}

        cpu_free = features.get("cpu_free", 0.5)
        if cpu_free > 0.7:
            params["threads"] = min(int(cpu_free * 8), 8)
            params["quality"] = "high"
        elif cpu_free > 0.4:
            params["threads"] = min(int(cpu_free * 4), 4)
            params["quality"] = "balanced"
        else:
            params["threads"] = 1
            params["quality"] = "fast"

        mem_free = features.get("mem_free_ratio", 0.5)
        if mem_free < 0.2:
            params["in_place"] = True
            params["buffer_size"] = "small"
        elif mem_free > 0.6:
            params["buffer_size"] = "large"

        time_budget = constraints.get("time_budget_ms", 500)
        if time_budget < 100:
            params["early_exit"] = True
            params["max_iterations"] = 100

        priority = constraints.get("priority", "balanced")
        if priority == "speed":
            params["approximate"] = True
            params["precision"] = "low"
        elif priority == "accuracy":
            params["approximate"] = False
            params["precision"] = "high"

        if hasattr(algo, "set_params"):
            algo.set_params(**params)

        return algo

    def _should_cache(self, algo: Any, context: Dict) -> bool:
        data_profile = context.get("data_profile", {})
        patterns = data_profile.get("patterns", {})

        if patterns.get("has_repeated_subsequences", False):
            return True
        if hasattr(algo, "call_count") and algo.call_count > 5:
            return True
        if data_profile.get("size", 0) < 1000:
            return True

        return False

    def _inject_memoization(self, algo: Any) -> Any:
        if not hasattr(algo, "process"):
            return algo

        original_process = algo.process
        cache_size = self.config.get("cache_size", 256)

        @functools.lru_cache(maxsize=cache_size)
        def cached_process(data_tuple):
            data = list(data_tuple) if isinstance(data_tuple, tuple) else data_tuple
            return original_process(data)

        def wrapper(data):
            if isinstance(data, list):
                data_tuple = tuple(data)
            else:
                data_tuple = data

            try:
                return cached_process(data_tuple)
            except (TypeError, HashError):
                return original_process(data)

        algo.process = wrapper
        algo.modifications.append("memoization")

        return algo

    def _should_parallelize(self, algo: Any, context: Dict) -> bool:
        if any(tag in algo.tags for tag in ["sorting", "stable"]):
            return False
        if hasattr(algo, 'name') and algo.name in ['timsort', 'merge_sort', 'heap_sort']:
            return False

        features = context.get("features", {})
        data_profile = context.get("data_profile", {})

        data_size = data_profile.get("size", 0)
        cpu_free = features.get("cpu_free", 0)
        cpu_count = features.get("cpu_count", 0)

        if data_size > 50000 and cpu_free > 0.5 and cpu_count > 1:
            return True

        predictions = context.get("predictions", {})
        if predictions.get("recommended_parallelism", 1) > 1:
            return True

        return False

    def _inject_parallelization(self, algo: Any, context: Dict) -> Any:
        if not hasattr(algo, "process"):
            return algo

        original_process = algo.process
        features = context.get("features", {})

        cpu_count = max(1, int(features.get("cpu_count", 1) * 16))
        parallelism = min(cpu_count - 1, 8)

        def parallel_process(data):
            if not isinstance(data, list) or len(data) < 1000:
                return original_process(data)

            chunk_size = max(1, len(data) // parallelism)
            chunks = [data[i:i + chunk_size] for i in range(0, len(data), chunk_size)]

            with ThreadPoolExecutor(max_workers=parallelism) as executor:
                results = list(executor.map(original_process, chunks))

            merged = []
            for r in results:
                if isinstance(r, list):
                    merged.extend(r)
                else:
                    merged.append(r)

            return merged

        algo.process = parallel_process
        algo.modifications.append(f"parallelization({parallelism})")

        return algo

    def _optimize_intermediate(self, algo: Any, context: Dict) -> Any:
        constraints = context.get("constraints", {})

        if constraints.get("priority") == "speed":
            if hasattr(algo, "set_params"):
                algo.set_params(streaming=True, buffer_size="small")

        return algo

    def _should_ast_optimize(self, algo: Any, context: Dict) -> bool:
        if hasattr(algo, "call_count") and algo.call_count > 20:
            return True
        predictions = context.get("predictions", {})
        if predictions.get("confidence", 0) > 0.8:
            return True
        return False

    def _ast_optimize(self, algo: Any, context: Dict) -> Any:
        try:
            if not hasattr(algo, "process"):
                return algo

            source = inspect.getsource(algo.process)
            source = textwrap.dedent(source)

            tree = ast.parse(source)

            optimizer = ASTOptimizer(context)
            tree = optimizer.visit(tree)
            ast.fix_missing_locations(tree)

            code = compile(tree, "<optimized>", "exec")
            namespace = {}
            exec(code, namespace)

            for name, obj in namespace.items():
                if callable(obj) and name != "__builtins__":
                    algo.process = types.MethodType(obj, algo)
                    algo.modifications.append("ast_optimization")
                    break

        except (OSError, TypeError, SyntaxError):
            pass

        return algo

    def execute_dag(self, graph: PipelineGraph, data: Any) -> Any:
        return graph.execute(data)

    def get_modification_stats(self) -> Dict[str, Any]:
        return {
            "total_modifications": len(self._modification_history),
            "modification_types": list(set(
                m["type"] for m in self._modification_history
            ))
        }


class ASTOptimizer(ast.NodeTransformer):
    def __init__(self, context: Dict):
        self.context = context
        self.optimizations_applied = []

    def visit_For(self, node: ast.For) -> ast.AST:
        self.generic_visit(node)

        if (len(node.body) == 1 and
            isinstance(node.body[0], ast.Expr) and
            isinstance(node.body[0].value, ast.Call)):

            call = node.body[0].value
            if (isinstance(call.func, ast.Attribute) and
                call.func.attr == "append"):
                self.optimizations_applied.append("list_comprehension")

        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        self.generic_visit(node)

        has_cache = any(
            isinstance(d, ast.Call) and
            isinstance(d.func, ast.Attribute) and
            d.func.attr == "lru_cache"
            for d in node.decorator_list
        )

        if not has_cache:
            cache_decorator = ast.parse("@functools.lru_cache(maxsize=128)").body[0].decorator_list[0]
            node.decorator_list.insert(0, cache_decorator)
            self.optimizations_applied.append("auto_cache")

        return node

    def visit_If(self, node: ast.If) -> ast.AST:
        self.generic_visit(node)
        return node


class HashError(Exception):
    pass



import heapq
from collections import deque
from algorithms.base import Algorithm


class Dijkstra(Algorithm):
    def __init__(self):
        super().__init__()
        self.name = "dijkstra"
        self.time_complexity = "O(V log V)"
        self.space_complexity = "O(V)"
        self.tags = ["pathfinding", "weighted", "shortest_path"]
        self.best_for = ["weighted_graphs", "non_negative_weights"]

    def validate_output(self, input_data, output_data):
        return isinstance(output_data, (list, set))

    def process(self, data):
        graph = data['graph']
        start = data['start']
        end = data.get('end')

        pq = [(0, start, [])]
        visited = set()

        while pq:
            cost, node, path = heapq.heappop(pq)
            if node in visited:
                continue
            visited.add(node)
            path = path + [node]

            if end and node == end:
                return path

            for neighbor, weight in graph.get(node, {}).items():
                if neighbor not in visited:
                    heapq.heappush(pq, (cost + weight, neighbor, path))

        return [] if end else visited


class AStar(Algorithm):
    def __init__(self):
        super().__init__()
        self.name = "a_star"
        self.time_complexity = "O(E)"
        self.space_complexity = "O(V)"
        self.tags = ["pathfinding", "heuristic", "grid"]
        self.best_for = ["grid_based", "geo_spatial"]

    def validate_output(self, input_data, output_data):
        return isinstance(output_data, list)

    def process(self, data):
        return Dijkstra().process(data)


class BFSPathfinder(Algorithm):
    def __init__(self):
        super().__init__()
        self.name = "bfs_path"
        self.time_complexity = "O(V+E)"
        self.space_complexity = "O(V)"
        self.tags = ["pathfinding", "unweighted", "breadth_first"]
        self.best_for = ["unweighted_graphs", "shortest_path"]

    def validate_output(self, input_data, output_data):
        return isinstance(output_data, list)

    def process(self, data):
        graph = data['graph']
        start = data['start']
        end = data['end']

        queue = deque([(start, [start])])
        visited = {start}

        while queue:
            node, path = queue.popleft()
            if node == end:
                return path

            for neighbor in graph.get(node, {}):
                if neighbor not in visited:
                    visited.add(neighbor)
                    queue.append((neighbor, path + [neighbor]))

        return []




from typing import Any, List, Optional, Dict
from algorithms.base import Algorithm
import numpy as np


class Primitive(Algorithm):
    """Base class for primitive operations used in algorithm synthesis."""

    time_complexity: str = "O(1)"
    space_complexity: str = "O(1)"
    best_for: List[str] = []
    combines_well_with: List[str] = []
    input_type: str = "any"
    output_type: str = "any"

    def __init__(self):
        super().__init__()
        self.transform_fn = lambda x: x.get('data') if isinstance(x, dict) else x

    @staticmethod
    def _unwrap_data(data):
        if isinstance(data, dict) and 'data' in data:
            return data['data']
        return data

    def can_compose_with(self, other: "Primitive") -> bool:
        return self.output_type == other.input_type or other.input_type == "any" or self.output_type == "any"

    def describe(self) -> Dict[str, Any]:
        info = super().describe()
        info.update({
            "time_complexity": self.time_complexity,
            "space_complexity": self.space_complexity,
            "best_for": self.best_for,
            "combines_well_with": self.combines_well_with,
            "input_type": self.input_type,
            "output_type": self.output_type
        })
        return info


class IteratePrimitive(Primitive):
    name = "iterate"
    tags = ["iteration", "loop", "basic"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "iterable"

    def process(self, data: Any) -> Any:
        if hasattr(data, '__iter__'):
            return list(data)
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if output_data is None:
            return False
        return True


class MapPrimitive(Primitive):
    name = "map"
    tags = ["iteration", "transformation", "functional"]
    time_complexity = "O(n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "iterable"
    combines_well_with = ["filter", "reduce"]

    def __init__(self, transform_fn=None):
        super().__init__()
        self.transform_fn = transform_fn

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.transform_fn:
            return [self.transform_fn(x) for x in data]
        if isinstance(data, list):
            return data
        return data


class FilterPrimitive(Primitive):
    name = "filter"
    tags = ["iteration", "selection", "functional"]
    time_complexity = "O(n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "iterable"
    combines_well_with = ["map", "reduce"]

    def __init__(self, predicate_fn=None):
        super().__init__()
        self.predicate_fn = predicate_fn

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.predicate_fn:
            return [x for x in data if self.predicate_fn(x)]
        if isinstance(data, list):
            return data
        return data


class ReducePrimitive(Primitive):
    name = "reduce"
    tags = ["iteration", "aggregation", "functional"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "scalar"
    combines_well_with = ["map", "filter"]

    def __init__(self, reduce_fn=None, initial=None):
        super().__init__()
        self.reduce_fn = reduce_fn
        self.initial = initial

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.reduce_fn:
            result = self.initial
            for x in data:
                result = self.reduce_fn(result, x) if result is not None else x
            return result
        if isinstance(data, list):
            return len(data)
        return data


class ScanPrimitive(Primitive):
    name = "scan"
    tags = ["iteration", "accumulation", "prefix"]
    time_complexity = "O(n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "iterable"
    combines_well_with = ["map", "filter"]

    def process(self, data: Any) -> Any:
        if isinstance(data, list):
            result = []
            acc = 0
            for x in data:
                if isinstance(x, (int, float)):
                    acc += x
                result.append(acc)
            return result
        return data


class PartitionPrimitive(Primitive):
    name = "partition"
    tags = ["iteration", "splitting", "divide"]
    time_complexity = "O(n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "tuple"
    combines_well_with = ["quicksort_primitive", "map"]

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and len(data) > 1:
            pivot = data[len(data) // 2]
            left = [x for x in data if x < pivot]
            middle = [x for x in data if x == pivot]
            right = [x for x in data if x > pivot]
            return (left, middle, right)
        return data


class BinarySearchPrimitive(Primitive):
    name = "binary_search"
    tags = ["search", "sorted", "logarithmic"]
    time_complexity = "O(log n)"
    space_complexity = "O(1)"
    input_type = "sorted_iterable"
    output_type = "scalar"
    best_for = ["sorted_data", "large_input"]
    combines_well_with = ["quicksort_primitive", "mergesort_primitive"]

    def __init__(self, target=None):
        super().__init__()
        self.target = target

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.target is not None:
            lo, hi = 0, len(data) - 1
            while lo <= hi:
                mid = (lo + hi) // 2
                if data[mid] == self.target:
                    return mid
                elif data[mid] < self.target:
                    lo = mid + 1
                else:
                    hi = mid - 1
            return -1
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if output_data is None:
            return False
        if isinstance(output_data, int) and self.target is not None:
            if output_data == -1:
                return self.target not in input_data
            return 0 <= output_data < len(input_data) and input_data[output_data] == self.target
        return True


class LinearSearchPrimitive(Primitive):
    name = "linear_search"
    tags = ["search", "unsorted", "linear"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["unsorted_data", "small_input"]

    def __init__(self, target=None):
        super().__init__()
        self.target = target

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.target is not None:
            for i, x in enumerate(data):
                if x == self.target:
                    return i
            return -1
        return data


class GreedyPrimitive(Primitive):
    name = "greedy"
    tags = ["optimization", "heuristic", "fast"]
    time_complexity = "O(n log n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "iterable"
    best_for = ["sorting_based", "selection"]
    combines_well_with = ["quicksort_primitive", "heapsort_primitive"]

    def __init__(self, select_fn=None):
        super().__init__()
        self.select_fn = select_fn

    def process(self, data: Any) -> Any:
        if isinstance(data, list):
            if self.select_fn:
                items = sorted(data, key=self.select_fn)
            else:
                items = sorted(data) if data else data
            return items
        return data


class DynamicProgrammingPrimitive(Primitive):
    name = "dynamic_programming"
    tags = ["optimization", "exact", "optimal"]
    time_complexity = "O(n²)"
    space_complexity = "O(n²)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["overlapping_subproblems", "optimal_substructure"]

    def __init__(self, dp_fn=None):
        super().__init__()
        self.dp_fn = dp_fn

    def process(self, data: Any) -> Any:
        if self.dp_fn and callable(self.dp_fn):
            return self.dp_fn(data)
        if isinstance(data, list) and len(data) > 0:
            return max(data)
        return data


class GradientDescentPrimitive(Primitive):
    name = "gradient_descent"
    tags = ["optimization", "iterative", "numerical"]
    time_complexity = "O(k*n)"
    space_complexity = "O(n)"
    input_type = "numeric_array"
    output_type = "scalar"
    best_for = ["convex", "continuous"]

    def __init__(self, learning_rate=0.01, max_iter=100):
        super().__init__()
        self.learning_rate = learning_rate
        self.max_iter = max_iter

    def process(self, data: Any) -> Any:
        if isinstance(data, (list, np.ndarray)) and len(data) > 0:
            nums = [float(x) for x in data if isinstance(x, (int, float))]
            if nums:
                result = sum(nums) / len(nums)
                return result
        return data


class QuickSortPrimitive(Primitive):
    name = "quicksort"
    tags = ["sorting", "divide_conquer", "in_place"]
    time_complexity = "O(n log n)"
    space_complexity = "O(log n)"
    input_type = "iterable"
    output_type = "sorted_iterable"
    best_for = ["general_sorting", "in_place_sort"]
    combines_well_with = ["binary_search", "interpolation_search", "greedy"]

    def process(self, data: Any) -> Any:
        if isinstance(data, list):
            if len(data) <= 1:
                return list(data)
            pivot = data[len(data) // 2]
            left = [x for x in data if x < pivot]
            middle = [x for x in data if x == pivot]
            right = [x for x in data if x > pivot]
            return self.process(left) + middle + self.process(right)
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if not isinstance(output_data, list):
            return False
        return all(output_data[i] <= output_data[i + 1] for i in range(len(output_data) - 1))


class MergeSortPrimitive(Primitive):
    name = "mergesort"
    tags = ["sorting", "divide_conquer", "stable"]
    time_complexity = "O(n log n)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "sorted_iterable"
    best_for = ["stable_sort", "external_sort", "large_data"]
    combines_well_with = ["binary_search", "interpolation_search"]

    def _merge(self, left, right):
        result = []
        i = j = 0
        while i < len(left) and j < len(right):
            if left[i] <= right[j]:
                result.append(left[i])
                i += 1
            else:
                result.append(right[j])
                j += 1
        result.extend(left[i:])
        result.extend(right[j:])
        return result

    def process(self, data: Any) -> Any:
        if isinstance(data, list):
            if len(data) <= 1:
                return list(data)
            mid = len(data) // 2
            left = self.process(data[:mid])
            right = self.process(data[mid:])
            return self._merge(left, right)
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if not isinstance(output_data, list):
            return False
        return all(output_data[i] <= output_data[i + 1] for i in range(len(output_data) - 1))


class HeapSortPrimitive(Primitive):
    name = "heapsort"
    tags = ["sorting", "heap", "selection"]
    time_complexity = "O(n log n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "sorted_iterable"
    best_for = ["in_place_sort", "priority_based"]
    combines_well_with = ["binary_search", "priority_queue"]

    def process(self, data: Any) -> Any:
        if isinstance(data, list):
            arr = list(data)
            n = len(arr)
            for i in range(n // 2 - 1, -1, -1):
                self._heapify(arr, n, i)
            for i in range(n - 1, 0, -1):
                arr[i], arr[0] = arr[0], arr[i]
                self._heapify(arr, i, 0)
            return arr
        return data

    def _heapify(self, arr, n, i):
        largest = i
        left = 2 * i + 1
        right = 2 * i + 2
        if left < n and arr[left] > arr[largest]:
            largest = left
        if right < n and arr[right] > arr[largest]:
            largest = right
        if largest != i:
            arr[i], arr[largest] = arr[largest], arr[i]
            self._heapify(arr, n, largest)

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if not isinstance(output_data, list):
            return False
        return all(output_data[i] <= output_data[i + 1] for i in range(len(output_data) - 1))


class BFSPrimitive(Primitive):
    name = "bfs"
    tags = ["graph", "traversal", "level_order"]
    time_complexity = "O(V + E)"
    space_complexity = "O(V)"
    input_type = "iterable"
    output_type = "iterable"
    best_for = ["graph_traversal", "shortest_path_unweighted", "level_order"]
    combines_well_with = ["dfs", "topological_sort"]

    def process(self, data: Any) -> Any:
        if isinstance(data, dict):
            start = next(iter(data.keys()), None)
            if start is None:
                return []
            visited = set()
            queue = [start]
            order = []
            visited.add(start)
            while queue:
                node = queue.pop(0)
                order.append(node)
                for neighbor in data.get(node, []):
                    if neighbor not in visited:
                        visited.add(neighbor)
                        queue.append(neighbor)
            return order
        if isinstance(data, list):
            return data
        return data


class DFSPrimitive(Primitive):
    name = "dfs"
    tags = ["graph", "traversal", "depth_first"]
    time_complexity = "O(V + E)"
    space_complexity = "O(V)"
    input_type = "iterable"
    output_type = "iterable"
    best_for = ["graph_traversal", "cycle_detection", "topological_sort"]
    combines_well_with = ["bfs", "topological_sort"]

    def process(self, data: Any) -> Any:
        if isinstance(data, dict):
            start = next(iter(data.keys()), None)
            if start is None:
                return []
            visited = set()
            order = []

            def _dfs(node):
                visited.add(node)
                order.append(node)
                for neighbor in data.get(node, []):
                    if neighbor not in visited:
                        _dfs(neighbor)

            _dfs(start)
            return order
        if isinstance(data, list):
            return data
        return data


class InterpolationSearchPrimitive(Primitive):
    name = "interpolation_search"
    tags = ["search", "sorted", "uniform"]
    time_complexity = "O(log log n)"
    space_complexity = "O(1)"
    input_type = "sorted_iterable"
    output_type = "scalar"
    best_for = ["uniformly_distributed", "large_sorted"]
    combines_well_with = ["quicksort", "mergesort", "heapsort"]

    def __init__(self, target=None):
        super().__init__()
        self.target = target

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.target is not None:
            lo, hi = 0, len(data) - 1
            while lo <= hi and data[lo] <= self.target <= data[hi]:
                if data[hi] == data[lo]:
                    if data[lo] == self.target:
                        return lo
                    break
                pos = lo + int((hi - lo) * (self.target - data[lo]) / (data[hi] - data[lo]))
                if data[pos] == self.target:
                    return pos
                if data[pos] < self.target:
                    lo = pos + 1
                else:
                    hi = pos - 1
            return -1
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if output_data is None:
            return False
        if isinstance(output_data, int) and self.target is not None:
            if output_data == -1:
                return True
            return 0 <= output_data < len(input_data) and input_data[output_data] == self.target
        return True


class TwoPointerPrimitive(Primitive):
    name = "two_pointer"
    tags = ["iteration", "linear", "pair_search"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "tuple"
    best_for = ["sorted_pairs", "palindrome", "window"]
    combines_well_with = ["quicksort", "mergesort", "partition"]

    def __init__(self, target_sum=None):
        super().__init__()
        self.target_sum = target_sum

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.target_sum is not None:
            sorted_data = sorted(data)
            left, right = 0, len(sorted_data) - 1
            while left < right:
                s = sorted_data[left] + sorted_data[right]
                if s == self.target_sum:
                    return (left, right, sorted_data[left], sorted_data[right])
                elif s < self.target_sum:
                    left += 1
                else:
                    right -= 1
            return None
        if isinstance(data, list):
            return (0, len(data) - 1)
        return data


class SlidingWindowPrimitive(Primitive):
    name = "sliding_window"
    tags = ["iteration", "window", "subarray"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["subarray", "substring", "contiguous"]
    combines_well_with = ["map", "reduce", "two_pointer"]

    def __init__(self, window_size=3):
        super().__init__()
        self.window_size = window_size

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and len(data) >= self.window_size:
            max_sum = sum(data[:self.window_size])
            current_sum = max_sum
            for i in range(self.window_size, len(data)):
                current_sum += data[i] - data[i - self.window_size]
                if current_sum > max_sum:
                    max_sum = current_sum
            return max_sum
        return data


class TopologicalSortPrimitive(Primitive):
    name = "topological_sort"
    tags = ["graph", "ordering", "dependency"]
    time_complexity = "O(V + E)"
    space_complexity = "O(V)"
    input_type = "iterable"
    output_type = "iterable"
    best_for = ["dependency_resolution", "scheduling", "ordering"]
    combines_well_with = ["bfs", "dfs", "greedy"]

    def process(self, data: Any) -> Any:
        if isinstance(data, dict):
            in_degree = {node: 0 for node in data}
            for node in data:
                for neighbor in data[node]:
                    in_degree[neighbor] = in_degree.get(neighbor, 0) + 1
            queue = [node for node, deg in in_degree.items() if deg == 0]
            result = []
            while queue:
                node = queue.pop(0)
                result.append(node)
                for neighbor in data.get(node, []):
                    in_degree[neighbor] -= 1
                    if in_degree[neighbor] == 0:
                        queue.append(neighbor)
            return result if len(result) == len(data) else []
        if isinstance(data, list):
            return data
        return data


class UnionFindPrimitive(Primitive):
    name = "union_find"
    tags = ["graph", "disjoint_set", "connectivity"]
    time_complexity = "O(α(n))"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["connectivity", "cycle_detection_graph", "kruskal"]
    combines_well_with = ["greedy", "kruskal"]

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and len(data) > 0:
            n = max(max(u, v) for u, v in data) if data and isinstance(data[0], (tuple, list)) else len(data)
            if isinstance(data[0], (tuple, list)):
                parent = list(range(n + 1))
                rank = [0] * (n + 1)

                def find(x):
                    while parent[x] != x:
                        parent[x] = parent[parent[x]]
                        x = parent[x]
                    return x

                def union(x, y):
                    rx, ry = find(x), find(y)
                    if rx == ry:
                        return False
                    if rank[rx] < rank[ry]:
                        parent[rx] = ry
                    elif rank[rx] > rank[ry]:
                        parent[ry] = rx
                    else:
                        parent[ry] = rx
                        rank[rx] += 1
                    return True

                components = n
                for u, v in data:
                    if union(u, v):
                        components -= 1
                return components
            return len(data)
        return data


class BacktrackingPrimitive(Primitive):
    name = "backtracking"
    tags = ["search", "recursive", "exhaustive"]
    time_complexity = "O(2ⁿ)"
    space_complexity = "O(n)"
    input_type = "iterable"
    output_type = "iterable"
    best_for = ["permutations", "combinations", "constraint_satisfaction"]
    combines_well_with = ["greedy", "branch_and_bound"]

    def __init__(self, backtrack_fn=None):
        super().__init__()
        self.backtrack_fn = backtrack_fn

    def process(self, data: Any) -> Any:
        if self.backtrack_fn and callable(self.backtrack_fn):
            return self.backtrack_fn(data)
        if isinstance(data, list):
            n = len(data)
            result = []

            def _permute(current, remaining):
                if not remaining:
                    result.append(list(current))
                for i in range(len(remaining)):
                    _permute(current + [remaining[i]], remaining[:i] + remaining[i + 1:])

            _permute([], data)
            return result
        return data


class RandomSearchPrimitive(Primitive):
    name = "random_search"
    tags = ["search", "heuristic", "stochastic"]
    time_complexity = "O(n)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["large_space", "no_structure", "exploration"]
    combines_well_with = ["greedy", "simulated_annealing"]

    def __init__(self, target=None, max_iterations=1000):
        super().__init__()
        self.target = target
        self.max_iterations = max_iterations

    def process(self, data: Any) -> Any:
        if isinstance(data, list) and self.target is not None:
            for _ in range(min(self.max_iterations, len(data))):
                import random
                idx = random.randint(0, len(data) - 1)
                if data[idx] == self.target:
                    return idx
            return -1
        return data


class LongestCommonSubsequencePrimitive(Primitive):
    name = "lcs"
    tags = ["string", "dynamic", "subsequence"]
    time_complexity = "O(n*m)"
    space_complexity = "O(n*m)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["sequence_matching", "diff", "bioinformatics"]
    combines_well_with = ["dynamic_programming", "string_match"]

    def process(self, data: Any) -> Any:
        if isinstance(data, (tuple, list)) and len(data) == 2:
            a, b = data[0], data[1]
            n, m = len(a), len(b)
            dp = [[0] * (m + 1) for _ in range(n + 1)]
            for i in range(1, n + 1):
                for j in range(1, m + 1):
                    if a[i - 1] == b[j - 1]:
                        dp[i][j] = dp[i - 1][j - 1] + 1
                    else:
                        dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])
            return dp[n][m]
        return data


class RabinKarpPrimitive(Primitive):
    name = "rabin_karp"
    tags = ["string", "pattern", "hashing"]
    time_complexity = "O(n + m)"
    space_complexity = "O(1)"
    input_type = "iterable"
    output_type = "scalar"
    best_for = ["pattern_matching", "string_search", "plagiarism"]
    combines_well_with = ["string_match", "lcs"]

    def __init__(self, pattern=None):
        super().__init__()
        self.pattern = pattern

    def process(self, data: Any) -> Any:
        if isinstance(data, str) and self.pattern:
            n, m = len(data), len(self.pattern)
            if m > n:
                return -1
            d, q = 256, 101
            p_hash = 0
            t_hash = 0
            h = pow(d, m - 1) % q
            for i in range(m):
                p_hash = (d * p_hash + ord(self.pattern[i])) % q
                t_hash = (d * t_hash + ord(data[i])) % q
            for i in range(n - m + 1):
                if p_hash == t_hash:
                    if data[i:i + m] == self.pattern:
                        return i
                if i < n - m:
                    t_hash = (d * (t_hash - ord(data[i]) * h) + ord(data[i + m])) % q
                    if t_hash < 0:
                        t_hash += q
            return -1
        return data

    def validate_output(self, input_data: Any, output_data: Any) -> bool:
        if output_data is None:
            return False
        if isinstance(output_data, int) and self.pattern is not None:
            if output_data == -1:
                return True
            return 0 <= output_data < len(input_data) and input_data[output_data:output_data + len(self.pattern)] == self.pattern
        return True


PRIMITIVES = {
    "iterate": IteratePrimitive(),
    "map": MapPrimitive(),
    "filter": FilterPrimitive(),
    "reduce": ReducePrimitive(),
    "scan": ScanPrimitive(),
    "partition": PartitionPrimitive(),
    "binary_search": BinarySearchPrimitive(),
    "linear_search": LinearSearchPrimitive(),
    "greedy": GreedyPrimitive(),
    "dynamic_programming": DynamicProgrammingPrimitive(),
    "gradient_descent": GradientDescentPrimitive(),
    "quicksort": QuickSortPrimitive(),
    "mergesort": MergeSortPrimitive(),
    "heapsort": HeapSortPrimitive(),
    "bfs": BFSPrimitive(),
    "dfs": DFSPrimitive(),
    "interpolation_search": InterpolationSearchPrimitive(),
    "two_pointer": TwoPointerPrimitive(),
    "sliding_window": SlidingWindowPrimitive(),
    "topological_sort": TopologicalSortPrimitive(),
    "union_find": UnionFindPrimitive(),
    "backtracking": BacktrackingPrimitive(),
    "random_search": RandomSearchPrimitive(),
    "lcs": LongestCommonSubsequencePrimitive(),
    "rabin_karp": RabinKarpPrimitive(),
}


def get_primitive_names() -> List[str]:
    return list(PRIMITIVES.keys())


def get_composable_chain(start: str, end: str) -> Optional[List[Primitive]]:
    if start not in PRIMITIVES or end not in PRIMITIVES:
        return None

    from collections import deque
    queue = deque()
    queue.append([PRIMITIVES[start]])
    visited = {start}

    while queue:
        path = queue.popleft()
        last = path[-1]

        if last.can_compose_with(PRIMITIVES[end]):
            return path + [PRIMITIVES[end]]

        for name, prim in PRIMITIVES.items():
            if name not in visited and last.can_compose_with(prim):
                visited.add(name)
                queue.append(path + [prim])

    return None


def compose_pipeline(primitive_names: List[str]) -> Optional[List[Primitive]]:
    pipeline = []
    for name in primitive_names:
        if name not in PRIMITIVES:
            return None
        pipeline.append(PRIMITIVES[name])

    for i in range(len(pipeline) - 1):
        if not pipeline[i].can_compose_with(pipeline[i + 1]):
            return None

    return pipeline



from .sorting_algos import (
    QuickSort,
    TimSort,
    HeapSort,
    InsertionSort,
    RadixSort,
    MergeSort
)

__all__ = [
    'QuickSort',
    'TimSort',
    'HeapSort',
    'InsertionSort',
    'RadixSort',
    'MergeSort'
]



import pytest
from algorithms.pathfinding import Dijkstra, AStar, BFSPathfinder


SIMPLE_GRAPH = {"A": {"B": 1, "C": 4}, "B": {"C": 2, "D": 5}, "C": {"D": 1}, "D": {}}


def test_dijkstra_finds_shortest_path():
    d = Dijkstra()
    result = d.process({"graph": SIMPLE_GRAPH, "start": "A", "end": "D"})
    assert result == ["A", "B", "C", "D"]


def test_dijkstra_no_end_returns_all_visited():
    d = Dijkstra()
    result = d.process({"graph": SIMPLE_GRAPH, "start": "A"})
    assert isinstance(result, set)
    assert result == {"A", "B", "C", "D"}


def test_dijkstra_no_path_returns_empty():
    d = Dijkstra()
    result = d.process({"graph": {"A": {}, "B": {}}, "start": "A", "end": "B"})
    assert result == []


def test_dijkstra_validate_output():
    d = Dijkstra()
    assert d.validate_output(None, ["A", "B"]) is True
    assert d.validate_output(None, {"A"}) is True
    assert d.validate_output(None, "invalid") is False


def test_astar_finds_path():
    a = AStar()
    result = a.process({"graph": SIMPLE_GRAPH, "start": "A", "end": "D"})
    assert result == ["A", "B", "C", "D"]


def test_astar_validate_output():
    a = AStar()
    assert a.validate_output(None, ["A", "B"]) is True
    assert a.validate_output(None, "invalid") is False


def test_bfs_pathfinder_finds_path():
    b = BFSPathfinder()
    result = b.process({"graph": SIMPLE_GRAPH, "start": "A", "end": "D"})
    assert result == ["A", "C", "D"] or result == ["A", "B", "D"]


def test_bfs_pathfinder_no_path_returns_empty():
    b = BFSPathfinder()
    result = b.process({"graph": {"A": {}, "B": {}}, "start": "A", "end": "B"})
    assert result == []


def test_bfs_validate_output():
    b = BFSPathfinder()
    assert b.validate_output(None, ["A"]) is True
    assert b.validate_output(None, None) is False


def test_dijkstra_tags():
    d = Dijkstra()
    assert "pathfinding" in d.tags
    assert "weighted" in d.tags


def test_astar_tags():
    a = AStar()
    assert "heuristic" in a.tags


def test_bfs_tags():
    b = BFSPathfinder()
    assert "unweighted" in b.tags


def test_data_auto_wrap_from_tuple():
    from core.problem_spec import ProblemSpec, ProblemType
    ps = ProblemSpec(name="test", problem_type=ProblemType.PATHFINDING)
    data = ps._infer_from_data_shape((SIMPLE_GRAPH, "A", "D"))
    assert data == ProblemType.PATHFINDING


def test_data_auto_wrap_from_dict_key():
    from core.problem_spec import ProblemSpec, ProblemType
    data = ProblemSpec._infer_from_data_shape({"graph": SIMPLE_GRAPH, "start": "A", "end": "D"})
    assert data == ProblemType.PATHFINDING



import unittest
import sys
import os
import json

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from interface.nl_parser import (
    parse_description, extract_data_from_description,
    parse_solve_input, PATTERNS
)
from core.problem_spec import ProblemSpec, ProblemType


class TestNLParseSorting(unittest.TestCase):
    def test_sort_numbers(self):
        spec = parse_description("sort these numbers in ascending order")
        self.assertEqual(spec.problem_type, ProblemType.TRANSFORMATION)
        self.assertIn("data", spec.inputs)

    def test_sort_description_name(self):
        desc = "sort this list of integers"
        spec = parse_description(desc)
        self.assertIn("sort", spec.name)

    def test_sort_extract_data(self):
        data = extract_data_from_description("sort numbers: 5, 3, 1, 4, 2")
        self.assertEqual(data, [5, 3, 1, 4, 2])


class TestNLParseSearch(unittest.TestCase):
    def test_find_element(self):
        spec = parse_description("find the number 42 in the list")
        self.assertEqual(spec.problem_type, ProblemType.SEARCH)

    def test_search_output_type(self):
        spec = parse_description("locate the target value in array")
        self.assertEqual(spec.outputs.get("result", {}).get("type"), "int")


class TestNLParseOptimization(unittest.TestCase):
    def test_maximize(self):
        spec = parse_description("maximize profit given constraints")
        self.assertEqual(spec.problem_type, ProblemType.OPTIMIZATION)

    def test_minimize_detection(self):
        spec = parse_description("minimize the total cost of the project")
        self.assertEqual(spec.problem_type, ProblemType.OPTIMIZATION)

    def test_direction_detected(self):
        spec = parse_description("maximize profit given constraints")
        self.assertGreater(len(spec.objectives), 0)


class TestNLParseRouting(unittest.TestCase):
    def test_routing_detection(self):
        spec = parse_description("find the shortest path between two cities on a map")
        self.assertEqual(spec.problem_type, ProblemType.ROUTING)
        self.assertGreater(len(spec.constraints), 0)


class TestNLParseScheduling(unittest.TestCase):
    def test_scheduling_detection(self):
        spec = parse_description("schedule these tasks to meet deadlines")
        self.assertEqual(spec.problem_type, ProblemType.SCHEDULING)


class TestNLParseClassification(unittest.TestCase):
    def test_classify_detection(self):
        spec = parse_description("classify these items into categories")
        self.assertEqual(spec.problem_type, ProblemType.CLASSIFICATION)

    def test_classify_output(self):
        spec = parse_description("classify these items into categories")
        self.assertEqual(spec.outputs.get("result", {}).get("type"), "int")


class TestNLParseInputInference(unittest.TestCase):
    def test_graph_input(self):
        spec = parse_description("find the shortest path in a graph with weighted edges")
        self.assertIn("graph", spec.inputs)

    def test_string_input(self):
        spec = parse_description("search for a pattern in a string")
        self.assertIn("text", spec.inputs)

    def test_fallback_input(self):
        spec = parse_description("do something with the data")
        self.assertIn("data", spec.inputs)


class TestNLParseSolveInput(unittest.TestCase):
    def test_parse_solve_input(self):
        spec, data = parse_solve_input("sort numbers: 10, 5, 2, 7")
        self.assertIsNotNone(spec)
        self.assertEqual(data, [10, 5, 2, 7])

    def test_parse_solve_no_numbers(self):
        spec, data = parse_solve_input("sort this list")
        self.assertIsNotNone(spec)
        self.assertIsNone(data)


class TestNLParseAllPatterns(unittest.TestCase):
    def test_all_patterns_have_required_keys(self):
        for p in PATTERNS:
            self.assertIn("name", p)
            self.assertIn("keywords", p)
            self.assertIn("problem_type", p)
            self.assertIn("output_type", p)
            self.assertGreater(len(p["keywords"]), 0)


class TestAPICreation(unittest.TestCase):
    def test_app_creation(self):
        try:
            from interface.api import create_app
            app = create_app()
            self.assertIsNotNone(app)
        except ImportError:
            self.skipTest("FastAPI not available")

    def test_app_solve_endpoint_exists(self):
        try:
            from interface.api import create_app
            app = create_app()
            routes = [r.path for r in app.routes]
            self.assertIn("/solve", routes)
        except ImportError:
            self.skipTest("FastAPI not available")

    def test_app_health_endpoint(self):
        try:
            from interface.api import create_app
            app = create_app()
            routes = [r.path for r in app.routes]
            self.assertIn("/health", routes)
        except ImportError:
            self.skipTest("FastAPI not available")


class TestCLICommands(unittest.TestCase):
    def test_cli_imports(self):
        try:
            from interface.cli import main
            self.assertIsNotNone(main)
        except ImportError:
            self.skipTest("CLI module issue")

    def test_subcommands_exist(self):
        try:
            from interface.cli import main
            commands = list(main.commands.keys())
            self.assertIn("solve", commands)
            self.assertIn("explain", commands)
            self.assertIn("stats", commands)
            self.assertIn("ml", commands)
        except ImportError:
            self.skipTest("CLI module issue")


class TestWebUIAvailability(unittest.TestCase):
    def test_import_gradio(self):
        try:
            from interface.web_ui import create_interface, launch
            self.assertTrue(True)
        except ImportError:
            self.skipTest("Gradio not available")


if __name__ == "__main__":
    unittest.main()



import json
import time
import logging
from typing import Any, Dict, List, Optional, Tuple, Callable
from dataclasses import dataclass, field

from core.problem_spec import ProblemSpec, ProblemType
from core.problem_library import ProblemLibrary
from algorithms.primitives import (
    PRIMITIVES, get_primitive_names, get_composable_chain,
    compose_pipeline, Primitive
)

logger = logging.getLogger(__name__)


@dataclass
class SynthesisResult:
    algorithms: List[Dict[str, Any]]
    strategy: str
    confidence: float
    explanation: str
    execution_time_ms: float = 0.0
    sources: List[str] = field(default_factory=list)
    pipeline: Optional[List[Primitive]] = None


class AlgorithmSynthesizer:
    def __init__(self, problem_library: Optional[ProblemLibrary] = None,
                 llm_client: Optional[Any] = None):
        self.problem_library = problem_library or ProblemLibrary()
        self.llm_client = llm_client
        self._synthesis_cache: Dict[str, SynthesisResult] = {}

    def synthesize(self, problem_spec: ProblemSpec,
                   use_llm: bool = False,
                   timeout_ms: int = 5000) -> SynthesisResult:
        start = time.perf_counter()
        cache_key = problem_spec.get_signature()

        if cache_key in self._synthesis_cache:
            cached = self._synthesis_cache[cache_key]
            cached.execution_time_ms = (time.perf_counter() - start) * 1000
            return cached

        result = self._cascade_synthesis(problem_spec, use_llm)

        result.execution_time_ms = (time.perf_counter() - start) * 1000
        self._synthesis_cache[cache_key] = result
        return result

    def _cascade_synthesis(self, spec: ProblemSpec, use_llm: bool) -> SynthesisResult:
        found, template_result = self._template_match(spec)
        if found:
            return template_result

        if self.problem_library and len(self.problem_library.problems) > 0:
            found, transfer_result = self._transfer_synthesis(spec)
            if found:
                return transfer_result

        if use_llm and self.llm_client:
            llm_result = self._llm_synthesis(spec)
            return llm_result

        composition_result = self._primitive_composition(spec)
        return composition_result

    def _template_match(self, spec: ProblemSpec) -> Tuple[bool, SynthesisResult]:
        algorithms = []
        pt = spec.problem_type

        if pt == ProblemType.OPTIMIZATION or pt == ProblemType.ROUTING:
            has_minimize = any(o.direction == "minimize" for o in spec.objectives)
            if spec.constraints:
                algorithms.append({
                    "name": "greedy",
                    "type": "template",
                    "confidence": 0.6,
                    "complexity": "O(n log n)"
                })
            algorithms.append({
                "name": "dynamic_programming",
                "type": "template",
                "confidence": 0.5,
                "complexity": "O(n²)"
            })

        elif pt == ProblemType.SEARCH:
            has_sorted = any("sorted" in str(c).lower() for c in spec.constraints)
            if has_sorted:
                algorithms.append({
                    "name": "binary_search",
                    "type": "template",
                    "confidence": 0.8,
                    "complexity": "O(log n)"
                })
            algorithms.append({
                "name": "linear_search",
                "type": "template",
                "confidence": 0.7,
                "complexity": "O(n)"
            })

        elif pt == ProblemType.TRANSFORMATION:
            algorithms.append({
                "name": "quicksort",
                "type": "template",
                "confidence": 0.7,
                "complexity": "O(n log n)"
            })
            algorithms.append({
                "name": "mergesort",
                "type": "template",
                "confidence": 0.7,
                "complexity": "O(n log n)"
            })

        elif pt == ProblemType.CLASSIFICATION:
            algorithms.append({
                "name": "greedy",
                "type": "template",
                "confidence": 0.5,
                "complexity": "O(n log n)"
            })

        elif pt == ProblemType.SCHEDULING:
            algorithms.append({
                "name": "topological_sort",
                "type": "template",
                "confidence": 0.6,
                "complexity": "O(V + E)"
            })
            algorithms.append({
                "name": "greedy",
                "type": "template",
                "confidence": 0.6,
                "complexity": "O(n log n)"
            })

        if not algorithms:
            return False, None

        return True, SynthesisResult(
            algorithms=algorithms,
            strategy="template_match",
            confidence=max(a["confidence"] for a in algorithms),
            explanation=f"Matched {pt.value} problem type with {len(algorithms)} template algorithm(s)",
            sources=["built-in templates"]
        )

    def _transfer_synthesis(self, spec: ProblemSpec) -> Tuple[bool, SynthesisResult]:
        if not self.problem_library or not self.problem_library.problems:
            return False, None

        similar = self.problem_library.find_similar(spec, top_k=3, min_similarity=0.3)
        if not similar:
            return False, None

        best_algos = self.problem_library.get_best_algorithms(spec, top_k=5)
        if not best_algos:
            return False, None

        algorithms = []
        sources = []
        max_conf = 0.0

        for algo_name, score in best_algos:
            if algo_name in PRIMITIVES:
                algorithms.append({
                    "name": algo_name,
                    "type": "transfer",
                    "confidence": score * 0.9,
                    "complexity": getattr(PRIMITIVES[algo_name], 'time_complexity', 'O(n)')
                })
                max_conf = max(max_conf, score * 0.9)

        for result in similar:
            sources.append(result.get("name", "unknown"))

        if not algorithms:
            return False, None

        return True, SynthesisResult(
            algorithms=algorithms,
            strategy="transfer",
            confidence=max_conf,
            explanation=f"Transferred from {len(sources)} similar problem(s): {', '.join(sources[:3])}",
            sources=sources
        )

    def _llm_synthesis(self, spec: ProblemSpec) -> SynthesisResult:
        prompt = self._build_llm_prompt(spec)
        code = ""
        explanation = ""

        if self.llm_client:
            try:
                response = self.llm_client.generate(prompt)
                code = response.get("code", "")
                explanation = response.get("explanation", "")
            except Exception as e:
                logger.warning(f"LLM synthesis failed: {e}")

        return SynthesisResult(
            algorithms=[{
                "name": "llm_generated",
                "type": "llm",
                "confidence": 0.4,
                "complexity": "unknown",
                "code": code
            }],
            strategy="llm",
            confidence=0.4,
            explanation=explanation or "Generated via LLM (low confidence, please validate)"
        )

    def _build_llm_prompt(self, spec: ProblemSpec) -> str:
        return json.dumps({
            "task": "Generate an algorithm for the given problem specification",
            "problem": spec.to_dict(),
            "available_primitives": list(PRIMITIVES.keys()),
            "output_format": {
                "code": "python function implementation",
                "explanation": "step-by-step explanation"
            }
        })

    def _primitive_composition(self, spec: ProblemSpec) -> SynthesisResult:
        pipeline = self._infer_pipeline(spec)
        algorithms = []

        if pipeline:
            pipeline_info = []
            for p in pipeline:
                pipeline_info.append({
                    "name": p.name,
                    "complexity": p.time_complexity
                })
            algorithms.append({
                "name": "composed_pipeline",
                "type": "composition",
                "confidence": 0.5,
                "pipeline": [p.name for p in pipeline],
                "complexities": [p.time_complexity for p in pipeline]
            })

        fallbacks = self._get_fallback_algorithms(spec)
        algorithms.extend(fallbacks)

        conf = max((a.get("confidence", 0) for a in algorithms), default=0.5)

        return SynthesisResult(
            algorithms=algorithms,
            strategy="primitive_composition",
            confidence=conf,
            explanation=f"Composed {len(algorithms)} algorithm candidate(s) from primitives",
            pipeline=pipeline
        )

    def _infer_pipeline(self, spec: ProblemSpec) -> Optional[List[Primitive]]:
        pt = spec.problem_type
        has_constraints = len(spec.constraints) > 0
        has_numeric_output = any(
            "int" in str(v.get("type", "")) or "float" in str(v.get("type", ""))
            for v in spec.outputs.values()
        )
        has_list_output = any(
            "list" in str(v.get("type", ""))
            for v in spec.outputs.values()
        )

        if pt == ProblemType.SEARCH:
            search_target_name = "search" if "search" in spec.name.lower() else None
            chain = get_composable_chain("iterate", "binary_search")
            if chain:
                return chain

        elif pt == ProblemType.OPTIMIZATION:
            if has_constraints:
                chain = get_composable_chain("iterate", "reduce")
                if chain:
                    return chain

        elif pt == ProblemType.TRANSFORMATION:
            if has_list_output:
                return compose_pipeline(["iterate", "map"])

        return None

    def _get_fallback_algorithms(self, spec: ProblemSpec) -> List[Dict]:
        algorithms = []

        input_types = set()
        for inp in spec.inputs.values():
            t = str(inp.get("type", "")).lower()
            if "list" in t:
                input_types.add("list")
            elif "int" in t or "float" in t:
                input_types.add("numeric")
            elif "str" in t:
                input_types.add("string")

        if "list" in input_types:
            algorithms.append({
                "name": "iterate",
                "type": "fallback",
                "confidence": 0.3,
                "complexity": "O(n)"
            })
            algorithms.append({
                "name": "map",
                "type": "fallback",
                "confidence": 0.3,
                "complexity": "O(n)"
            })
            if spec.problem_type == ProblemType.SEARCH:
                algorithms.append({
                    "name": "linear_search",
                    "type": "fallback",
                    "confidence": 0.4,
                    "complexity": "O(n)"
                })

        if "string" in input_types and spec.problem_type == ProblemType.SEARCH:
            algorithms.append({
                "name": "rabin_karp",
                "type": "fallback",
                "confidence": 0.3,
                "complexity": "O(n+m)"
            })

        return algorithms

    def clear_cache(self):
        self._synthesis_cache.clear()




import time
import psutil
import numpy as np
from typing import Any, Dict, List, Optional, Tuple
import json

class ContextEngine:
    """
    AI-powered context analysis engine.
    Profiles input data, environment, and predicts optimal conditions.
    """

    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}
        self.history: List[Dict] = []
        self._cache: Dict[str, Any] = {}

    def analyze(self, data: Any, task_type: str = "auto") -> Dict[str, Any]:
        """
        Comprehensive context analysis with AI-enhanced feature extraction.
        """
        context = {
            "timestamp": time.time(),
            "task_type": task_type,
            "data_profile": self._analyze_data(data),
            "environment": self._analyze_environment(),
            "constraints": self._analyze_constraints(),
            "predictions": {},
            "features": {}
        }

        # AI-enhanced predictions
        context["predictions"] = self._predict_optimal_conditions(context)
        context["features"] = self._extract_features(context)

        if context["task_type"] == "unknown" or context["task_type"] == "auto":
            if hasattr(data, '__len__'):
                context["task_type"] = "safety"
            else:
                context["task_type"] = "safety"
        known_types = {"sorting", "pathfinding", "optimization", "clustering", "classification", "regression", "search", "safety", "transformation", "image_processing"}
        if context["task_type"] not in known_types:
            context["task_type"] = "safety"

        self.history.append(context)
        return context

    def _analyze_data(self, data: Any) -> Dict[str, Any]:
        """Deep data profiling."""
        profile = {
            "type": type(data).__name__,
            "size": None,
            "shape": None,
            "memory_bytes": None,
            "statistics": {},
            "patterns": {}
        }

        if hasattr(data, "__len__"):
            profile["size"] = len(data)

        if hasattr(data, "shape"):
            profile["shape"] = data.shape

        if hasattr(data, "nbytes"):
            profile["memory_bytes"] = data.nbytes
        elif isinstance(data, (list, tuple, set)):
            profile["memory_bytes"] = self._estimate_memory(data)

        # Statistical analysis for numeric data
        profile["statistics"] = self._compute_statistics(data)

        # Pattern detection
        profile["patterns"] = self._detect_patterns(data)

        return profile

    def _compute_statistics(self, data: Any) -> Dict[str, Any]:
        """Compute statistical properties of data."""
        stats = {}

        try:
            if isinstance(data, (list, tuple)) and len(data) > 0:
                if all(isinstance(x, (int, float)) for x in data[:100]):
                    arr = np.array(data, dtype=float)
                    stats["mean"] = float(np.mean(arr))
                    stats["std"] = float(np.std(arr))
                    stats["min"] = float(np.min(arr))
                    stats["max"] = float(np.max(arr))
                    stats["range"] = stats["max"] - stats["min"]
                    stats["skewness"] = float(self._compute_skewness(arr))
                    stats["is_uniform"] = stats["std"] < (stats["range"] * 0.1) if stats["range"] > 0 else False
                    stats["has_duplicates"] = len(set(data)) < len(data)
                    stats["unique_ratio"] = len(set(data)) / len(data)

        except Exception as e:
            stats["error"] = str(e)

        return stats

    def _compute_skewness(self, arr: np.ndarray) -> float:
        """Compute skewness of distribution."""
        if len(arr) < 3:
            return 0.0
        mean = np.mean(arr)
        std = np.std(arr)
        if std == 0:
            return 0.0
        return float(np.mean(((arr - mean) / std) ** 3))

    def _detect_patterns(self, data: Any) -> Dict[str, Any]:
        """Detect structural patterns in data."""
        patterns = {}

        if isinstance(data, list) and len(data) > 1:
            try:
                patterns["is_sorted"] = all(data[i] <= data[i+1] for i in range(min(len(data)-1, 1000)))
                patterns["is_reverse_sorted"] = all(data[i] >= data[i+1] for i in range(min(len(data)-1, 1000)))
            except TypeError:
                patterns["is_sorted"] = False
                patterns["is_reverse_sorted"] = False

            # Check for nearly sorted (few inversions)
            if len(data) <= 10000:
                try:
                    inversions = sum(1 for i in range(min(len(data)-1, 5000)) for j in range(i+1, min(len(data), i+100)) if data[i] > data[j])
                except TypeError:
                    inversions = len(data)
                patterns["inversion_count"] = inversions
                patterns["is_nearly_sorted"] = inversions < len(data) * 0.1
            else:
                # Sample for large data
                sample = data[::len(data)//1000 + 1]
                inversions = sum(1 for i in range(len(sample)-1) for j in range(i+1, min(len(sample), i+50)) if sample[i] > sample[j])
                patterns["is_nearly_sorted"] = inversions < len(sample) * 0.1

            # Check for repeated patterns
            if len(data) > 100:
                patterns["has_repeated_subsequences"] = self._check_repeated_patterns(data)

        return patterns

    def _check_repeated_patterns(self, data: list) -> bool:
        """Check if data has repeated subsequences."""
        if len(data) < 20:
            return False
        window = min(10, len(data) // 10)
        seen = set()
        for i in range(0, min(len(data) - window, 1000), window):
            sub = tuple(data[i:i+window])
            if sub in seen:
                return True
            seen.add(sub)
        return False

    def _analyze_environment(self) -> Dict[str, Any]:
        """Analyze system environment."""
        env = {
            "cpu": {
                "percent_used": psutil.cpu_percent(interval=0.1),
                "count": psutil.cpu_count(),
                "freq_mhz": None
            },
            "memory": {
                "total_gb": psutil.virtual_memory().total / (1024**3),
                "available_gb": psutil.virtual_memory().available / (1024**3),
                "percent_used": psutil.virtual_memory().percent
            },
            "disk": {
                "free_gb": psutil.disk_usage('/').free / (1024**3)
            }
        }

        try:
            env["cpu"]["freq_mhz"] = psutil.cpu_freq().current if psutil.cpu_freq() else None
        except:
            pass

        return env

    def _analyze_constraints(self) -> Dict[str, Any]:
        """Analyze current constraints."""
        return {
            "time_budget_ms": self.config.get("time_budget_ms", 500),
            "memory_budget_mb": self.config.get("memory_budget_mb", 1024),
            "accuracy_target": self.config.get("accuracy_target", 0.95),
            "priority": self.config.get("priority", "balanced")  # speed, accuracy, balanced
        }

    def _predict_optimal_conditions(self, context: Dict) -> Dict[str, Any]:
        """AI-powered prediction of optimal execution conditions."""
        predictions = {
            "recommended_parallelism": 1,
            "recommended_batch_size": None,
            "risk_factors": [],
            "confidence": 0.5
        }

        env = context["environment"]
        data = context["data_profile"]

        # Predict parallelism based on CPU availability
        cpu_free = 100 - env["cpu"]["percent_used"]
        cpu_count = env["cpu"]["count"]

        if cpu_free > 70 and cpu_count > 2:
            predictions["recommended_parallelism"] = min(cpu_count - 1, 8)
        elif cpu_free > 40 and cpu_count > 1:
            predictions["recommended_parallelism"] = min(cpu_count // 2, 4)

        # Predict batch size based on memory
        mem_free_gb = env["memory"]["available_gb"]
        if mem_free_gb > 8:
            predictions["recommended_batch_size"] = "large"
        elif mem_free_gb > 2:
            predictions["recommended_batch_size"] = "medium"
        else:
            predictions["recommended_batch_size"] = "small"
            predictions["risk_factors"].append("low_memory")

        # Risk assessment
        if env["memory"]["percent_used"] > 90:
            predictions["risk_factors"].append("memory_pressure")
        if cpu_free < 20:
            predictions["risk_factors"].append("cpu_contention")

        # Confidence based on data quality
        if data["size"] is not None and data["size"] > 100:
            predictions["confidence"] = 0.8
        else:
            predictions["confidence"] = 0.6

        return predictions

    def _extract_features(self, context: Dict) -> Dict[str, float]:
        """Extract normalized feature vector for ML models."""
        features = {}

        data = context["data_profile"]
        env = context["environment"]

        # Data features
        features["data_size_log"] = np.log10(data["size"] + 1) if data["size"] else 0
        features["is_numeric"] = 1.0 if data["type"] in ("list", "tuple", "ndarray") else 0.0
        features["is_sorted"] = 1.0 if data.get("patterns", {}).get("is_sorted", False) else 0.0
        features["is_nearly_sorted"] = 1.0 if data.get("patterns", {}).get("is_nearly_sorted", False) else 0.0

        # Environment features
        features["cpu_free"] = (100 - env["cpu"]["percent_used"]) / 100.0
        features["mem_free_ratio"] = (100 - env["memory"]["percent_used"]) / 100.0
        features["cpu_count"] = min(env["cpu"]["count"], 16) / 16.0

        # Constraint features
        constraints = context["constraints"]
        features["time_budget_norm"] = min(constraints["time_budget_ms"] / 1000.0, 1.0)
        features["priority_speed"] = 1.0 if constraints["priority"] == "speed" else 0.0
        features["priority_accuracy"] = 1.0 if constraints["priority"] == "accuracy" else 0.0

        return features

    def _estimate_memory(self, data: list) -> int:
        """Estimate memory usage of a list."""
        try:
            import sys
            return sys.getsizeof(data)
        except:
            return len(data) * 8  # rough estimate

    def get_feature_vector(self, context: Dict) -> List[float]:
        """Get flat feature vector for ML consumption."""
        return list(context["features"].values())



"""Test the pre-trained RL model across all domains."""

import sys
import os
import time
import random
import numpy as np

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from core.rl import PPOAgent
from core.context_engine import ContextEngine
from core.problem_spec import ProblemSpec, ProblemType
from pipeline import UniversalSolver

solver = UniversalSolver()
context_engine = ContextEngine()
registry = solver._get_global_registry()
algo_names = list(registry.keys())
algo_to_idx = {n: i for i, n in enumerate(algo_names)}
idx_to_algo = {i: n for n, i in algo_to_idx.items()}


def build_state(data, problem_type):
    context = context_engine.analyze(data, problem_type.value)
    vec = np.zeros(200, dtype=np.float32)
    dp = context.get("data_profile", {})
    patterns = dp.get("patterns", {})
    vec[0] = np.log10(dp.get("size", 1) + 1) / 5.0
    vec[1] = 1.0 if patterns.get("is_sorted") else 0.0
    vec[2] = 1.0 if patterns.get("is_nearly_sorted") else 0.0
    env_info = context.get("environment", {})
    vec[10] = (100.0 - env_info.get("cpu", {}).get("percent_used", 50)) / 100.0
    all_types = list(ProblemType)
    if problem_type in all_types:
        idx = all_types.index(problem_type)
        if 20 + idx < 200:
            vec[20 + idx] = 1.0
    return vec


def eval_algo(algo_name, data):
    algo = registry[algo_name]
    try:
        start = time.time()
        result = algo.process(data)
        elapsed = time.time() - start
        valid = algo.validate_output(data, result)
        return valid, elapsed * 1000, result
    except Exception as e:
        return False, 0, str(e)


def run_domain_test(domain, test_cases):
    print(f"\n{'='*60}")
    print(f"Domain: {domain.upper()}")
    print(f"{'='*60}")
    for spec, data, desc in test_cases:
        state = build_state(data, spec.problem_type)
        action, _, _ = agent.select_action(state, deterministic=True)
        algo_name = idx_to_algo[action % len(idx_to_algo)]
        valid, ms, _ = eval_algo(algo_name, data)
        status = "PASS" if valid else "FAIL"
        print(f"  [{status}] {desc:40s} -> {algo_name:20s} ({ms:6.1f}ms)")


def generate_graph(n, d=0.3):
    nodes = [str(i) for i in range(n)]
    graph = {n: {} for n in nodes}
    for i in range(n):
        for j in range(i + 1, min(i + 3, n)):
            if random.random() < d:
                graph[nodes[i]][nodes[j]] = random.randint(1, 10)
    return {'graph': graph, 'start': nodes[0], 'end': nodes[-1]}


if __name__ == "__main__":
    for label, ckpt_path in [
        ("Final Model", "checkpoints/pretrain_v2/pretrained_final.pt"),
        ("Best Ep1500", "checkpoints/pretrain_v2/best_model_ep1500.pt"),
    ]:
        if not os.path.exists(ckpt_path):
            print(f"Checkpoint not found: {ckpt_path}")
            continue

        agent = PPOAgent(state_dim=200, num_actions=len(algo_names), config={})
        agent.load(ckpt_path)
        agent.network.eval()

        print(f"\n{'='*70}")
        print(f"  {label}")
        print(f"{'='*70}")

        run_domain_test("sorting", [
            (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
             [random.randint(0, 1000) for _ in range(100)], "100 random ints"),
            (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
             list(range(1000)), "1K sorted"),
            (ProblemSpec(name="s", problem_type=ProblemType.TRANSFORMATION),
             list(range(10000, 0, -1)), "10K reverse"),
        ])

        run_domain_test("pathfinding", [
            (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
             generate_graph(10, 0.4), "10-node sparse graph"),
            (ProblemSpec(name="p", problem_type=ProblemType.PATHFINDING),
             generate_graph(26, 0.3), "26-node graph"),
        ])



"""
Distributed Training (Multi-GPU)

Speeds up training ~10x across multiple GPUs.

Usage:
    torchrun --nproc_per_node=4 training/distributed_train.py
"""

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
import os
import logging

logger = logging.getLogger(__name__)


def setup_distributed(rank: int, world_size: int):
    os.environ["MASTER_ADDR"] = os.environ.get("MASTER_ADDR", "localhost")
    os.environ["MASTER_PORT"] = os.environ.get("MASTER_PORT", "12355")
    dist.init_process_group("nccl" if torch.cuda.is_available() else "gloo",
                            rank=rank, world_size=world_size)


def cleanup_distributed():
    dist.destroy_process_group()


def compute_loss(policy: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
    """Simplified loss for demonstration."""
    targets = torch.zeros_like(policy)
    targets[:, 0] = 1.0
    ce = torch.nn.functional.cross_entropy(policy, targets.argmax(dim=-1))
    return ce


def train_distributed(rank: int, world_size: int):
    setup_distributed(rank, world_size)
    logger.info("Rank %d / %d: online", rank, world_size)

    from core.rl.agents.selection_agent import PPOAgent

    agent = PPOAgent(state_dim=200, num_actions=20)
    device = torch.device(f"cuda:{rank}" if torch.cuda.is_available() else "cpu")
    agent.network = agent.network.to(device)
    agent.network = torch.nn.parallel.DistributedDataParallel(
        agent.network,
        device_ids=[rank] if torch.cuda.is_available() else None,
    )

    for iteration in range(10000):
        state = torch.randn(64, 200).to(device)
        policy, value = agent.network(state)
        loss = compute_loss(policy, value)
        agent.optimizer.zero_grad()
        loss.backward()
        agent.optimizer.step()

        if rank == 0 and (iteration + 1) % 100 == 0:
            logger.info("Iteration %d, Loss: %.4f", iteration + 1, loss.item())

    if rank == 0:
        import os as _os
        _os.makedirs("checkpoints", exist_ok=True)
        torch.save(agent.network.module.state_dict(), "checkpoints/distributed_final.pt")
        logger.info("Saved distributed_final.pt")

    cleanup_distributed()


def main():
    if torch.cuda.is_available():
        world_size = torch.cuda.device_count()
    else:
        world_size = 2
        logger.warning("No CUDA — using %d CPU processes for demo", world_size)

    mp.spawn(train_distributed, args=(world_size,), nprocs=world_size, join=True)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
    main()
